# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

from typing import Any, Dict, FrozenSet, List, NamedTuple, Optional, Set, Union

import graphene
from graphql.execution.base import ResolveInfo
from sqlalchemy import func
from sqlalchemy.orm import Session, aliased

from ..filter import Filter
from ..models import (
    DBID,
    Issue,
    IssueInstance,
    IssueInstanceSharedTextAssoc,
    IssueStatus,
    SharedText,
    SharedTextKind,
    SourceLocation,
)
from ..queries import get_warning_message
from ..sarif_types import SARIFSeverityLevel, SARIFResult
from . import filter_predicates, run

# pyre-fixme[5]: Global expression must be annotated.
FilenameText = aliased(SharedText)
# pyre-fixme[5]: Global expression must be annotated.
CallableText = aliased(SharedText)
# pyre-fixme[5]: Global expression must be annotated.
CallerText = aliased(SharedText)
# pyre-fixme[5]: Global expression must be annotated.
CalleeText = aliased(SharedText)
# pyre-fixme[5]: Global expression must be annotated.
MessageText = aliased(SharedText)
# pyre-fixme[5]: Global expression must be annotated.
FeatureText = aliased(SharedText)
# pyre-fixme[5]: Global expression must be annotated.
SourceNameText = aliased(SharedText)
# pyre-fixme[5]: Global expression must be annotated.
SourceKindText = aliased(SharedText)
# pyre-fixme[5]: Global expression must be annotated.
SinkNameText = aliased(SharedText)
# pyre-fixme[5]: Global expression must be annotated.
SinkKindText = aliased(SharedText)


# pyre-ignore[13]: unitialized class attribute
class IssueQueryResultType(graphene.ObjectType):
    concatenated_features: str

    class Meta:
        interfaces = (graphene.relay.Node,)

    issue_id = graphene.ID()
    issue_instance_id = graphene.ID()

    code = graphene.Int()
    message = graphene.String()

    callable = graphene.String()
    status = graphene.String()

    filename = graphene.String()
    location = graphene.String()

    sources = graphene.List(graphene.String)
    source_names = graphene.List(graphene.String)
    sinks = graphene.List(graphene.String)
    sink_names = graphene.List(graphene.String)
    features = graphene.List(graphene.String)

    is_new_issue = graphene.Boolean()
    first_seen = graphene.String()

    min_trace_length_to_sources = graphene.Int()
    min_trace_length_to_sinks = graphene.Int()

    warning_message = graphene.String()

    def resolve_sources(self, info: ResolveInfo) -> List[str]:
        # pyre-ignore[6]: graphene too dynamic.
        return list(sources(info.context["session"], self.issue_instance_id))

    def resolve_source_names(self, info: ResolveInfo) -> List[str]:
        # pyre-ignore[6]: graphene too dynamic.
        return list(source_names(info.context["session"], self.issue_instance_id))

    def resolve_sinks(self, info: ResolveInfo) -> List[str]:
        # pyre-ignore[6]: graphene too dynamic.
        return list(sinks(info.context["session"], self.issue_instance_id))

    def resolve_sink_names(self, info: ResolveInfo) -> List[str]:
        # pyre-ignore[6]: graphene too dynamic.
        return list(sink_names(info.context["session"], self.issue_instance_id))

    def resolve_features(self, info: ResolveInfo) -> List[str]:
        # pyre-ignore[6]: graphene too dynamic.
        return sorted(self.features)

    def resolve_warning_message(self, info: ResolveInfo) -> str:
        # pyre-ignore[6]: graphene too dynamic.
        warning_message = get_warning_message(info.context["session"], self.code)
        if warning_message:
            return warning_message.message
        return ""


class IssueQueryResult(NamedTuple):
    issue_id: DBID
    issue_instance_id: DBID

    code: int
    message: str

    callable: str
    status: str

    filename: str
    location: SourceLocation

    first_seen: str
    is_new_issue: bool

    min_trace_length_to_sources: int
    min_trace_length_to_sinks: int

    features: FrozenSet[str]

    source_names: FrozenSet[str]
    source_kinds: FrozenSet[str]
    sink_names: FrozenSet[str]
    sink_kinds: FrozenSet[str]

    @staticmethod
    # pyre-fixme[2]: Parameter annotation cannot be `Any`.
    def from_record(record: Any) -> IssueQueryResult:
        return IssueQueryResult(
            issue_id=record.issue_id,
            issue_instance_id=record.issue_instance_id,
            code=record.code,
            message=record.message,
            callable=record.callable,
            status=record.status.name.replace("_", " ").capitalize(),
            first_seen=record.first_seen,
            filename=record.filename,
            location=record.location,
            is_new_issue=record.is_new_issue,
            min_trace_length_to_sources=record.min_trace_length_to_sources,
            min_trace_length_to_sinks=record.min_trace_length_to_sinks,
            features=frozenset(record.concatenated_features.split(","))
            if record.concatenated_features
            else frozenset(),
            source_names=frozenset(record.concatenated_source_names.split(","))
            if record.concatenated_source_names
            else frozenset(),
            source_kinds=frozenset(record.concatenated_source_kinds.split(","))
            if record.concatenated_source_kinds
            else frozenset(),
            sink_names=frozenset(record.concatenated_sink_names.split(","))
            if record.concatenated_sink_names
            else frozenset(),
            sink_kinds=frozenset(record.concatenated_sink_kinds.split(","))
            if record.concatenated_sink_kinds
            else frozenset(),
        )

    def to_json(self) -> Dict[str, Union[str, int, List[str], bool]]:
        return {
            "issue_id": self.issue_id.resolved(),
            "filename": self.filename,
            "line": self.location.line_no,
            "begin_column": self.location.begin_column,
            "end_column": self.location.end_column,
            "code": self.code,
            "message": self.message,
            "callable": self.callable,
            "status": self.status,
            "source_names": list(self.source_names),
            "source_kinds": list(self.source_kinds),
            "sink_names": list(self.sink_names),
            "sink_kinds": list(self.sink_kinds),
            "min_trace_length_to_sources": self.min_trace_length_to_sources,
            "min_trace_length_to_sinks": self.min_trace_length_to_sinks,
            "features": list(self.features),
            "is_new_issue": self.is_new_issue,
            "first_seen": self.first_seen,
        }

    def to_sarif(self, severity_level: str = "warning") -> SARIFResult:
        sarif_result = {
            "ruleId": str(self.code),
            "level": str(SARIFSeverityLevel(severity_level)),
            "message": {
                "text": self.message,
            },
            "locations": [
                {
                    "physicalLocation": {
                        "artifactLocation": {"uri": self.filename},
                        "region": self.location.to_sarif(),
                    }
                }
            ],
        }
        return sarif_result

    def __hash__(self) -> int:
        return hash(
            (
                self.issue_id.resolved(),
                self.issue_instance_id.resolved(),
                self.code,
                self.message,
                self.callable,
                self.status,
                self.source_names,
                self.source_kinds,
                self.sink_names,
                self.sink_kinds,
                self.filename,
                self.location,
                self.is_new_issue,
                self.first_seen,
                self.min_trace_length_to_sinks,
                self.min_trace_length_to_sources,
                self.features,
            )
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return (
            self.issue_id.resolved() == other.issue_id.resolved()
            and self.issue_instance_id.resolved() == other.issue_instance_id.resolved()
            and self.code == other.code
            and self.status == other.status
            and self.message == other.message
            and self.callable == other.callable
            and self.source_names == other.source_names
            and self.source_kinds == other.source_kinds
            and self.sink_names == other.sink_names
            and self.sink_kinds == other.sink_kinds
            and self.filename == other.filename
            and self.location == other.location
            and self.is_new_issue == other.is_new_issue
            and self.first_seen == other.first_seen
            and self.min_trace_length_to_sinks == other.min_trace_length_to_sinks
            and self.min_trace_length_to_sources == other.min_trace_length_to_sources
            and self.features == other.features
        )


class Instance:
    def __init__(self, session: Session, run_id: Optional[DBID] = None) -> None:
        self._session: Session = session
        self._predicates: List[filter_predicates.Predicate] = []
        self._run_id: DBID = run_id or run.latest(session)

    def get(self) -> List[IssueQueryResult]:
        features = (
            self._session.query(
                # pyre-ignore: SQAlchemy sadness.
                IssueInstance.id.label("id"),
                func.group_concat(FeatureText.contents.distinct()).label(
                    "concatenated_features"
                ),
            )
            .join(
                IssueInstanceSharedTextAssoc,
                IssueInstanceSharedTextAssoc.issue_instance_id == IssueInstance.id,
                isouter=True,
            )
            .join(
                FeatureText,
                FeatureText.id == IssueInstanceSharedTextAssoc.shared_text_id,
                isouter=True,
            )
            .filter(FeatureText.kind == SharedTextKind.FEATURE)
            .group_by(IssueInstance)
            .subquery()
        )
        source_names = (
            self._session.query(
                IssueInstance.id.label("id"),
                func.group_concat(SourceNameText.contents.distinct()).label(
                    "concatenated_source_names"
                ),
            )
            .join(
                IssueInstanceSharedTextAssoc,
                IssueInstanceSharedTextAssoc.issue_instance_id == IssueInstance.id,
            )
            .join(
                SourceNameText,
                SourceNameText.id == IssueInstanceSharedTextAssoc.shared_text_id,
            )
            .filter(SourceNameText.kind == SharedTextKind.SOURCE_DETAIL)
            .group_by(IssueInstance)
            .subquery()
        )
        source_kinds = (
            self._session.query(
                IssueInstance.id.label("id"),
                func.group_concat(SourceKindText.contents.distinct()).label(
                    "concatenated_source_kinds"
                ),
            )
            .join(
                IssueInstanceSharedTextAssoc,
                IssueInstanceSharedTextAssoc.issue_instance_id == IssueInstance.id,
            )
            .join(
                SourceKindText,
                SourceKindText.id == IssueInstanceSharedTextAssoc.shared_text_id,
            )
            .filter(SourceKindText.kind == SharedTextKind.SOURCE)
            .group_by(IssueInstance)
            .subquery()
        )
        sink_names = (
            self._session.query(
                IssueInstance.id.label("id"),
                func.group_concat(SinkNameText.contents.distinct()).label(
                    "concatenated_sink_names"
                ),
            )
            .join(
                IssueInstanceSharedTextAssoc,
                IssueInstanceSharedTextAssoc.issue_instance_id == IssueInstance.id,
            )
            .join(
                SinkNameText,
                SinkNameText.id == IssueInstanceSharedTextAssoc.shared_text_id,
            )
            .filter(SinkNameText.kind == SharedTextKind.SINK_DETAIL)
            .group_by(IssueInstance)
            .subquery()
        )
        sink_kinds = (
            self._session.query(
                IssueInstance.id.label("id"),
                func.group_concat(SinkKindText.contents.distinct()).label(
                    "concatenated_sink_kinds"
                ),
            )
            .join(
                IssueInstanceSharedTextAssoc,
                IssueInstanceSharedTextAssoc.issue_instance_id == IssueInstance.id,
            )
            .join(
                SinkKindText,
                SinkKindText.id == IssueInstanceSharedTextAssoc.shared_text_id,
            )
            .filter(SinkKindText.kind == SharedTextKind.SINK)
            .group_by(IssueInstance)
            .subquery()
        )

        query = (
            self._session.query(
                IssueInstance.id.label("issue_instance_id"),
                FilenameText.contents.label("filename"),
                IssueInstance.location,
                # pyre-ignore[16]: SQLAlchemy
                Issue.id.label("issue_id"),
                Issue.code,
                Issue.status,
                Issue.first_seen,
                CallableText.contents.label("callable"),
                MessageText.contents.label("message"),
                IssueInstance.is_new_issue,
                IssueInstance.min_trace_length_to_sources,
                IssueInstance.min_trace_length_to_sinks,
                features.c.concatenated_features,
                source_names.c.concatenated_source_names,
                source_kinds.c.concatenated_source_kinds,
                sink_names.c.concatenated_sink_names,
                sink_kinds.c.concatenated_sink_kinds,
            )
            .filter(IssueInstance.run_id == self._run_id)
            .join(FilenameText, FilenameText.id == IssueInstance.filename_id)
            .join(CallableText, CallableText.id == IssueInstance.callable_id)
            .join(features, IssueInstance.id == features.c.id, isouter=True)
            .join(source_names, IssueInstance.id == source_names.c.id, isouter=True)
            .join(source_kinds, IssueInstance.id == source_kinds.c.id, isouter=True)
            .join(sink_names, IssueInstance.id == sink_names.c.id, isouter=True)
            .join(sink_kinds, IssueInstance.id == sink_kinds.c.id, isouter=True)
        )

        for predicate in self._predicates:
            if isinstance(predicate, filter_predicates.QueryPredicate):
                query = predicate.apply(query)

        issues = [
            IssueQueryResult.from_record(record)
            for record in query.join(Issue, IssueInstance.issue_id == Issue.id).join(
                MessageText, MessageText.id == IssueInstance.message_id
            )
        ]

        issue_predicates = [
            predicate
            for predicate in self._predicates
            if isinstance(predicate, filter_predicates.IssuePredicate)
        ]
        if len(issue_predicates) > 0:
            for issue_predicate in issue_predicates:
                issues = issue_predicate.apply(issues)

        return issues

    def where(self, *predicates: filter_predicates.Predicate) -> "Instance":
        self._predicates.extend(predicates)
        return self

    def where_issue_instance_id_is(self, issue_id: Optional[int]) -> "Instance":
        if issue_id is not None:
            self._predicates.append(
                filter_predicates.Equals(IssueInstance.id, issue_id)
            )
        return self

    def where_is_new_issue(self, is_new_issue: Optional[bool]) -> "Instance":
        if is_new_issue:
            self._predicates.append(
                filter_predicates.Equals(IssueInstance.is_new_issue, True)
            )
        return self

    def where_codes_is_any_of(self, codes: List[int]) -> "Instance":
        return self.where(filter_predicates.Like(Issue.code, codes))

    def where_callables_is_any_of(self, callables: List[str]) -> "Instance":
        return self.where(filter_predicates.Like(CallableText.contents, callables))

    def where_callables_matches(self, regex: str) -> "Instance":
        return self.where(filter_predicates.Matches(regex, "callable"))

    def where_status_is_any_of(self, statuses: List[str]) -> "Instance":
        return self.where(filter_predicates.Like(Issue.status, statuses))

    def where_path_is_any_of(self, paths: List[str]) -> "Instance":
        return self.where(filter_predicates.Like(FilenameText.contents, paths))

    def where_source_name_is_any_of(self, source_names: List[str]) -> "Instance":
        return self.where(filter_predicates.HasAny(set(source_names), "source_names"))

    def where_source_name_matches(self, regex: str) -> "Instance":
        return self.where(filter_predicates.Matches(regex, "source_names"))

    def where_source_kind_is_any_of(self, source_kinds: List[str]) -> "Instance":
        return self.where(filter_predicates.HasAny(set(source_kinds), "source_kinds"))

    def where_source_kind_matches(self, regex: str) -> "Instance":
        return self.where(filter_predicates.Matches(regex, "source_kinds"))

    def where_sink_name_is_any_of(self, sink_names: List[str]) -> "Instance":
        return self.where(filter_predicates.HasAny(set(sink_names), "sink_names"))

    def where_sink_name_matches(self, regex: str) -> "Instance":
        return self.where(filter_predicates.Matches(regex, "sink_names"))

    def where_sink_kind_is_any_of(self, sink_kinds: List[str]) -> "Instance":
        return self.where(filter_predicates.HasAny(set(sink_kinds), "sink_kinds"))

    def where_sink_kind_matches(self, regex: str) -> "Instance":
        return self.where(filter_predicates.Matches(regex, "sink_kinds"))

    def where_trace_length_to_sinks(
        self, minimum: Optional[int] = None, maximum: Optional[int] = None
    ) -> "Instance":
        return self.where(
            filter_predicates.InRange(
                IssueInstance.min_trace_length_to_sinks, lower=minimum, upper=maximum
            )
        )

    def where_filter(self, filter_instance: Filter) -> "Instance":
        traceLengthToSinks = filter_instance.traceLengthToSinks
        min_trace_length_to_sinks: Optional[int] = None
        if traceLengthToSinks is not None:
            min_trace_length_to_sinks = traceLengthToSinks[0]

        max_trace_length_to_sinks: Optional[int] = None
        if traceLengthToSinks is not None:
            max_trace_length_to_sinks = traceLengthToSinks[1]

        traceLengthFromSources = filter_instance.traceLengthFromSources
        min_trace_length_to_sources: Optional[int] = None
        if traceLengthFromSources is not None:
            min_trace_length_to_sources = traceLengthFromSources[0]

        max_trace_length_to_sources: Optional[int] = None
        if traceLengthFromSources is not None:
            max_trace_length_to_sources = traceLengthFromSources[1]

        builder = (
            self.where_codes_is_any_of(filter_instance.codes)
            .where_path_is_any_of(filter_instance.paths)
            .where_trace_length_to_sinks(
                min_trace_length_to_sinks, max_trace_length_to_sinks
            )
            .where_trace_length_to_sources(
                min_trace_length_to_sources, max_trace_length_to_sources
            )
            .where_is_new_issue(filter_instance.is_new_issue)
            .where_status_is_any_of(filter_instance.statuses)
        )

        if filter_instance.callables:
            if not isinstance(filter_instance.callables, List):
                # pyre-fixme[16]: `List` has no attribute 'get'
                if filter_instance.callables.get("operation") == "matches":
                    builder = builder.where_callables_matches(
                        filter_instance.callables.get("value", [""])[0]
                    )
                elif filter_instance.callables.get("operation") == "is":
                    if any(filter_instance.callables.get("value", [])):
                        builder = builder.where_callables_is_any_of(
                            filter_instance.callables.get("value", [""])
                        )
                else:
                    raise ValueError(
                        """Invalid value supplied for callables parameter
                        `operation`. The supported values are `is` and
                        `matches`."""
                    )
            else:
                # Backward compatibility
                # pyre-fixme[6]: Expected `List[str]` for 1st positional only
                # parameter but got
                # `Union[Dict[str, Union[List[str], str]], List[str]]`
                builder = builder.where_callables_is_any_of(filter_instance.callables)

        if filter_instance.source_names:
            if not isinstance(filter_instance.source_names, list):
                if filter_instance.source_names.get("operation") == "matches":
                    builder = builder.where_source_name_matches(
                        filter_instance.source_names.get("value", [""])[0]
                    )
                elif filter_instance.source_names.get("operation") == "is":
                    if any(filter_instance.source_names.get("value", [])):
                        builder = builder.where_source_name_is_any_of(
                            filter_instance.source_names.get("value", [""])
                        )
                else:
                    raise ValueError(
                        """Invalid value supplied for source_names parameter
                        `operation`. The supported values are `is` and
                        `matches`."""
                    )
            else:
                # Backward compatibility
                builder = builder.where_source_name_is_any_of(
                    # pyre-fixme[6]: Expected `List[str]` for 1st positional only
                    # parameter but got
                    # `Union[Dict[str, Union[List[str], str]], List[str]]`
                    filter_instance.source_names
                )

        if filter_instance.source_kinds:
            if not isinstance(filter_instance.source_kinds, list):
                if filter_instance.source_kinds.get("operation") == "matches":
                    builder = builder.where_source_kind_matches(
                        filter_instance.source_kinds.get("value", [""])[0]
                    )
                elif filter_instance.source_kinds.get("operation") == "is":
                    if any(filter_instance.source_kinds.get("value", [])):
                        builder = builder.where_source_kind_is_any_of(
                            filter_instance.source_kinds.get("value", [""])
                        )
                else:
                    raise ValueError(
                        """Invalid value supplied for source_kinds parameter
                        `operation`. The supported values are `is` and
                        `matches`."""
                    )
            else:
                # Backward compatibility
                builder = builder.where_source_kind_is_any_of(
                    # pyre-fixme[6]: Expected `List[str]` for 1st positional only
                    # parameter but got
                    # `Union[Dict[str, Union[List[str], str]], List[str]]`
                    filter_instance.source_kinds
                )

        if filter_instance.sink_names:
            if not isinstance(filter_instance.sink_names, list):
                if filter_instance.sink_names.get("operation") == "matches":
                    builder = builder.where_sink_name_matches(
                        filter_instance.sink_names.get("value", [""])[0]
                    )
                elif filter_instance.sink_names.get("operation") == "is":
                    if any(filter_instance.sink_names.get("value", [])):
                        builder = builder.where_sink_name_is_any_of(
                            filter_instance.sink_names.get("value", [""])
                        )
                else:
                    raise ValueError(
                        """Invalid value supplied for sink_names parameter
                        `operation`. The supported values are `is` and
                        `matches`."""
                    )
            else:
                # Backward compatibility
                # pyre-fixme[6]: Expected `List[str]` for 1st positional only
                # parameter but got
                # `Union[Dict[str, Union[List[str], str]], List[str]]`
                builder = builder.where_sink_name_is_any_of(filter_instance.sink_names)

        if filter_instance.sink_kinds:
            if not isinstance(filter_instance.sink_kinds, list):
                if filter_instance.sink_kinds.get("operation") == "matches":
                    builder = builder.where_sink_kind_matches(
                        filter_instance.sink_kinds.get("value", [""])[0]
                    )
                elif filter_instance.sink_kinds.get("operation") == "is":
                    if any(filter_instance.sink_kinds.get("value", [])):
                        builder = builder.where_sink_kind_is_any_of(
                            filter_instance.sink_kinds.get("value", [""])
                        )
                else:
                    raise ValueError(
                        """Invalid value supplied for sink_kinds parameter
                        `operation`. The supported values are `is` and
                        `matches`."""
                    )
            else:
                # Backward compatibility if the filter is of the form List[str]
                # pyre-fixme[6]: Expected `List[str]` for 1st positional only
                # parameter but got
                # `Union[Dict[str, Union[List[str], str]], List[str]]`
                builder = builder.where_sink_kind_is_any_of(filter_instance.sink_kinds)

        for feature in filter_instance.format_features_for_query() or []:
            if feature[0] == "any of":
                builder = builder.where_any_features(feature[1])
            if feature[0] == "all of":
                builder = builder.where_all_features(feature[1])
            if feature[0] == "none of":
                builder = builder.where_exclude_features(feature[1])

        return builder

    def where_trace_length_to_sources(
        self, minimum: Optional[int] = None, maximum: Optional[int] = None
    ) -> "Instance":
        return self.where(
            filter_predicates.InRange(
                IssueInstance.min_trace_length_to_sources, lower=minimum, upper=maximum
            )
        )

    def where_any_features(self, features: List[str]) -> "Instance":
        return self.where(filter_predicates.HasAny(set(features), "features"))

    def where_all_features(self, features: List[str]) -> "Instance":
        return self.where(filter_predicates.HasAll(set(features)))

    def where_exclude_features(self, features: List[str]) -> "Instance":
        return self.where(filter_predicates.HasNone(set(features)))


def sources(session: Session, issue_id: DBID) -> Set[str]:
    return _get_leaves(session, issue_id, SharedTextKind.SOURCE)


def source_names(session: Session, issue_id: DBID) -> Set[str]:
    return _get_leaves(session, issue_id, SharedTextKind.SOURCE_DETAIL)


def sinks(session: Session, issue_id: DBID) -> Set[str]:
    return _get_leaves(session, issue_id, SharedTextKind.SINK)


def sink_names(session: Session, issue_id: DBID) -> Set[str]:
    return _get_leaves(session, issue_id, SharedTextKind.SINK_DETAIL)


def features(session: Session, issue_id: DBID) -> Set[str]:
    return _get_leaves(session, issue_id, SharedTextKind.FEATURE)


def _get_leaves(
    session: Session, issue_instance_id: DBID, kind: SharedTextKind
) -> Set[str]:
    message_ids = [
        int(id)
        for id, in session.query(SharedText.id)
        .distinct(SharedText.id)
        .join(
            IssueInstanceSharedTextAssoc,
            SharedText.id == IssueInstanceSharedTextAssoc.shared_text_id,
        )
        .filter(IssueInstanceSharedTextAssoc.issue_instance_id == issue_instance_id)
        .filter(SharedText.kind == kind)
    ]

    leaf_lookup = {
        int(id): contents
        for id, contents in session.query(SharedText.id, SharedText.contents).filter(
            SharedText.kind == kind
        )
    }
    return {leaf_lookup[id] for id in message_ids if id in leaf_lookup}


def update_status(session: Session, id: str, status: str) -> None:
    status_enums = {
        "bad_practice": IssueStatus.BAD_PRACTICE,
        "false_positive": IssueStatus.FALSE_POSITIVE,
        "valid_bug": IssueStatus.VALID_BUG,
        "do_not_care": IssueStatus.DO_NOT_CARE,
        "uncategorized": IssueStatus.UNCATEGORIZED,
    }
    session.query(Issue).filter(Issue.id == id).update({"status": status_enums[status]})
    session.commit()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import re
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Generic,
    List,
    Optional,
    Pattern,
    Sequence,
    Set,
    TypeVar,
    Union,
)

from sqlalchemy import Column
from sqlalchemy.orm.query import Query
from sqlalchemy.sql.expression import or_
from typing_extensions import Final

from ..models import DBID

if TYPE_CHECKING:
    from .issues import IssueQueryResult  # noqa

_Q = TypeVar("_Q")
_T = TypeVar("_T")


class Predicate(ABC):
    pass


class QueryPredicate(Predicate):
    @abstractmethod
    def apply(self, query: Query[_Q]) -> Query[_Q]:
        ...


class InRange(Generic[_T], QueryPredicate):
    def __init__(
        self,
        column: Union[Column[_T], DBID],
        *,
        lower: Optional[_T] = None,
        upper: Optional[_T] = None,
    ) -> None:
        self._column = column
        self._lower: Final[Optional[_T]] = lower
        self._upper: Final[Optional[_T]] = upper

    def apply(self, query: Query[_Q]) -> Query[_Q]:
        if self._lower is not None:
            query = query.filter(self._column >= self._lower)
        if self._upper is not None:
            query = query.filter(self._column <= self._upper)
        return query


class Equals(Generic[_T], QueryPredicate):
    def __init__(self, column: Union[Column[_T], DBID], to: _T) -> None:
        self._column = column
        self._to: Final[Optional[_T]] = to

    def apply(self, query: Query[_Q]) -> Query[_Q]:
        return query.filter(self._column == self._to)


class IsNull(Generic[_T], QueryPredicate):
    def __init__(self, column: Union[Column[_T], DBID]) -> None:
        self._column = column

    def apply(self, query: Query[_Q]) -> Query[_Q]:
        return query.filter(self._column is None)


class Like(Generic[_T], QueryPredicate):
    def __init__(self, column: Union[Column[_T], DBID], items: Sequence[_T]) -> None:
        self._column = column
        self._items = items

    def apply(self, query: Query[_Q]) -> Query[_Q]:
        # pyre-ignore: SQLAlchemy too dynamic.
        return query.filter(or_(*[self._column.like(item) for item in self._items]))


class IssuePredicate(Predicate):
    @abstractmethod
    def apply(self, issues: List[IssueQueryResult]) -> List[IssueQueryResult]:
        ...


class HasAll(IssuePredicate):
    def __init__(self, features: Set[str]) -> None:
        self._features = features

    def apply(self, issues: List[IssueQueryResult]) -> List[IssueQueryResult]:
        return [
            issue
            for issue in issues
            if issue.features & self._features == self._features
        ]


class Matches(IssuePredicate):
    def __init__(self, regex: str, parameter_name: str) -> None:
        self._regex: Pattern[str] = re.compile(regex)
        self._parameter_name = parameter_name

    def attribute_set(self, issue: IssueQueryResult) -> Set[str]:
        attribute = issue._asdict()[self._parameter_name]
        if isinstance(attribute, str):
            return {attribute}
        return set(attribute)

    def apply(self, issues: List[IssueQueryResult]) -> List[IssueQueryResult]:
        return [
            issue
            for issue in issues
            if any(map(self._regex.match, self.attribute_set(issue)))
        ]


class HasAny(IssuePredicate):
    def __init__(self, parameter_list: Set[str], parameter_name: str) -> None:
        self._parameter_list = parameter_list
        self._parameter_name = parameter_name

    def attribute_set(self, issue: IssueQueryResult) -> Set[str]:
        attribute = issue._asdict()[self._parameter_name]
        if isinstance(attribute, str):
            return {attribute}
        return set(attribute)

    def apply(self, issues: List[IssueQueryResult]) -> List[IssueQueryResult]:
        return [
            issue
            for issue in issues
            if not self.attribute_set(issue).isdisjoint(self._parameter_list)
        ]


class HasNone(IssuePredicate):
    def __init__(self, features: Set[str]) -> None:
        self._features = features

    def apply(self, issues: List[IssueQueryResult]) -> List[IssueQueryResult]:
        return [issue for issue in issues if len(issue.features & self._features) == 0]



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import sys
from abc import ABCMeta, abstractmethod
from datetime import datetime, timedelta
from enum import Enum
from typing import (
    Any,
    Dict,
    Generic,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from ..sarif_types import SARIFRegionObject


if sys.version_info >= (3, 8):
    from typing import TypedDict, Literal
else:
    from typing_extensions import TypedDict, Literal


# pyre-fixme[5]: Global expression must be annotated.
log = logging.getLogger("sapp")

T = TypeVar("T")
T_in = TypeVar("T_in")
T_out = TypeVar("T_out")


class ParseError(Exception):
    def __init__(self, message: str, received: object = None) -> None:
        self.message = message
        self.received = received

    def __str__(self) -> str:
        message = self.message
        if self.received is not None:
            message = f"{message}\nReceived: `{self.received}`"
        return message


class ParseType(Enum):
    ISSUE = "issue"
    PRECONDITION = "precondition"
    POSTCONDITION = "postcondition"


# NB: The TypedDict types are an approximation of the legacy untyped maps emitted
# by the various parsers. They are transformed into the NamedTuple versions immediately
# after parsing, before the rest of the pipeline is executed.
# This is for performance reasons (we intern very common strings, and NamedTuples themselves are more memory-efficient).

# Parsers can also choose to directly return the new NamedTuple types.
# Eventually, we should convert all the parsers to return the NamedTuple versions,
# and delete the TypedDict versions. But it's a ton of work to do so
# (especially converting all the unit tests), so we have two versions for now.


class ParsePosition(TypedDict, total=False):
    filename: str
    line: int
    start: int
    end: int


class SourceLocation(NamedTuple):
    """The location in a source file that an error occurred in

    If end_column is defined then we have a range, otherwise it defaults to
    begin_column and we have a single point.
    """

    line_no: int
    begin_column: int
    end_column: int

    @staticmethod
    def of(
        line_no: int, begin_column: int, end_column: Optional[int] = None
    ) -> "SourceLocation":
        return SourceLocation(line_no, begin_column, end_column or begin_column)

    def __str__(self) -> str:
        return SourceLocation.to_string(self)

    def to_sarif(self) -> SARIFRegionObject:
        region = {
            "startLine": self.line_no,
            "startColumn": self.begin_column,
        }
        if self.end_column:
            region["endColumn"] = self.end_column + 1
        return region

    @staticmethod
    def from_typed_dict(d: ParsePosition) -> "SourceLocation":
        return SourceLocation(
            d["line"],
            d["start"],
            d["end"],
        )

    @staticmethod
    def from_string(location_string: str) -> "SourceLocation":
        location_points = location_string.split("|")
        assert len(location_points) == 3, "Invalid location string %s" % location_string
        return SourceLocation(*map(int, location_points))

    @staticmethod
    def to_string(location: "SourceLocation") -> str:
        return "|".join(
            map(str, [location.line_no, location.begin_column, location.end_column])
        )


class ParseTypeInterval(NamedTuple):
    start: int
    finish: int
    preserves_type_context: bool


class ParseTraceAnnotation(NamedTuple):
    location: SourceLocation
    kind: str
    msg: str
    leaf_kind: Optional[str]
    leaf_depth: int
    type_interval: Optional[ParseTypeInterval]
    link: Optional[str]
    trace_key: Optional[str]
    titos: List[SourceLocation]
    subtraces: List[Dict[str, Any]]  # TODO figure what exactly this shape is

    @staticmethod
    def from_json(j: Dict[str, Any]) -> "ParseTraceAnnotation":
        return ParseTraceAnnotation(
            location=SourceLocation.from_typed_dict(j["location"]),
            kind=j["kind"],
            msg=j["msg"],
            leaf_kind=j.get("leaf_kind"),
            leaf_depth=j["leaf_depth"],
            type_interval=j.get("type_interval"),
            link=j.get("link"),
            trace_key=j.get("trace_key"),
            titos=list(map(SourceLocation.from_typed_dict, j.get("titos", []))),
            subtraces=j["subtraces"],
        )


ParseFeature = Dict[str, str]
ParseLeaf = Tuple[str, int]  # (kind, distance)
ParseIssueLeaf = Tuple[str, str, int]  # (callable, kind, distance)


def flatten_feature(feature: ParseFeature) -> Iterable[str]:
    for key, value in feature.items():
        if isinstance(value, str) and value:
            yield key + ":" + value
        else:
            yield key


def flatten_features(features: Iterable[ParseFeature]) -> List[str]:
    ret = []
    for feature in features:
        ret.extend(flatten_feature(feature))
    return ret


def intern_leaves(leaves: Iterable[ParseLeaf]) -> List[ParseLeaf]:
    return list(map(lambda p: (sys.intern(p[0]), p[1]), leaves))


class ParseCondition(TypedDict, total=False):
    type: Union[Literal[ParseType.PRECONDITION], Literal[ParseType.POSTCONDITION]]
    callable: str
    caller: str
    caller_port: str
    filename: str
    callee: str
    callee_port: str
    callee_location: ParsePosition
    leaves: Iterable[ParseLeaf]
    type_interval: Optional[ParseTypeInterval]
    features: Iterable[ParseFeature]
    titos: Iterable[ParsePosition]
    annotations: Iterable[Dict[str, Any]]


class ParseConditionTuple(NamedTuple):
    type: Union[Literal[ParseType.PRECONDITION], Literal[ParseType.POSTCONDITION]]
    caller: str
    caller_port: str
    filename: str
    callee: str
    callee_port: str
    callee_location: SourceLocation
    leaves: List[ParseLeaf]
    type_interval: Optional[ParseTypeInterval]
    features: List[str]
    titos: Iterable[SourceLocation]
    annotations: Iterable[ParseTraceAnnotation]

    @staticmethod
    def from_typed_dict(d: ParseCondition) -> "ParseConditionTuple":
        return ParseConditionTuple(
            type=d["type"],
            caller=d["caller"],
            caller_port=d["caller_port"],
            filename=d["filename"],
            callee=d["callee"],
            callee_port=d["callee_port"],
            callee_location=SourceLocation.from_typed_dict(d["callee_location"]),
            leaves=list(d["leaves"]),
            type_interval=d["type_interval"],
            features=flatten_features(d.get("features", [])),
            titos=list(map(SourceLocation.from_typed_dict, d.get("titos", []))),
            annotations=list(
                map(ParseTraceAnnotation.from_json, d.get("annotations", []))
            ),
        )

    def interned(self) -> "ParseConditionTuple":
        "Return self, but with certain strings interned"
        return ParseConditionTuple(
            type=self.type,
            caller=self.caller,
            caller_port=sys.intern(self.caller_port),
            filename=self.filename,
            callee=self.callee,
            callee_port=sys.intern(self.callee_port),
            callee_location=self.callee_location,
            leaves=intern_leaves(self.leaves),
            type_interval=self.type_interval,
            features=list(map(sys.intern, self.features)),
            titos=self.titos,
            annotations=self.annotations,
        )


class ParseIssueCondition(TypedDict):
    callee: str
    port: str
    location: ParsePosition
    leaves: Iterable[ParseLeaf]
    titos: Iterable[ParsePosition]
    features: Iterable[ParseFeature]
    type_interval: Optional[ParseTypeInterval]
    annotations: Iterable[Dict[str, Any]]


class ParseIssueConditionTuple(NamedTuple):
    callee: str
    port: str
    location: SourceLocation
    leaves: List[ParseLeaf]
    titos: Iterable[SourceLocation]
    features: List[str]
    type_interval: Optional[ParseTypeInterval]
    annotations: Iterable[ParseTraceAnnotation]

    @staticmethod
    def from_typed_dict(d: ParseIssueCondition) -> "ParseIssueConditionTuple":
        return ParseIssueConditionTuple(
            callee=d["callee"],
            port=d["port"],
            location=SourceLocation.from_typed_dict(d["location"]),
            leaves=list(d["leaves"]),
            titos=list(map(SourceLocation.from_typed_dict, d.get("titos", []))),
            features=flatten_features(d.get("features", [])),
            type_interval=d["type_interval"],
            annotations=list(
                map(ParseTraceAnnotation.from_json, d.get("annotations", []))
            ),
        )

    def interned(self) -> "ParseIssueConditionTuple":
        "Return self, but with certain strings interned"
        return ParseIssueConditionTuple(
            callee=self.callee,
            port=sys.intern(self.port),
            location=self.location,
            leaves=intern_leaves(self.leaves),
            titos=self.titos,
            features=list(map(sys.intern, self.features)),
            type_interval=self.type_interval,
            annotations=self.annotations,
        )


class ParseIssue(TypedDict, total=False):
    type: Literal[ParseType.ISSUE]
    code: int
    message: str
    callable: str
    handle: str
    filename: str
    callable_line: int
    line: int
    start: int
    end: int
    preconditions: Iterable[ParseIssueCondition]
    postconditions: Iterable[ParseIssueCondition]
    initial_sources: Iterable[ParseIssueLeaf]
    final_sinks: Iterable[ParseIssueLeaf]
    features: Iterable[ParseFeature]
    fix_info: Optional[Dict[str, Any]]


class ParseIssueTuple(NamedTuple):
    code: int
    message: str
    callable: str
    handle: str
    filename: str
    line: int
    start: int
    end: int
    preconditions: Iterable[ParseIssueConditionTuple]
    postconditions: Iterable[ParseIssueConditionTuple]
    initial_sources: Iterable[ParseIssueLeaf]
    final_sinks: Iterable[ParseIssueLeaf]
    features: List[str]
    callable_line: Optional[int]
    fix_info: Optional[Dict[str, Any]]

    @staticmethod
    def from_typed_dict(d: ParseIssue) -> "ParseIssueTuple":
        return ParseIssueTuple(
            code=d["code"],
            message=d["message"],
            callable=d["callable"],
            handle=d["handle"],
            filename=d["filename"],
            callable_line=d.get("callable_line", None),
            line=d["line"],
            start=d["start"],
            end=d["end"],
            preconditions=list(
                map(ParseIssueConditionTuple.from_typed_dict, d["preconditions"])
            ),
            postconditions=list(
                map(ParseIssueConditionTuple.from_typed_dict, d["postconditions"])
            ),
            initial_sources=d["initial_sources"],
            final_sinks=d["final_sinks"],
            features=flatten_features(d["features"]),
            fix_info=d.get("fix_info"),
        )

    def interned(self) -> "ParseIssueTuple":
        return ParseIssueTuple(
            code=self.code,
            message=self.message,
            callable=self.callable,
            handle=self.handle,
            filename=self.filename,
            callable_line=self.callable_line,
            line=self.line,
            start=self.start,
            end=self.end,
            preconditions=list(
                map(ParseIssueConditionTuple.interned, self.preconditions)
            ),
            postconditions=list(
                map(ParseIssueConditionTuple.interned, self.postconditions)
            ),
            initial_sources=self.initial_sources,
            final_sinks=self.final_sinks,
            features=list(map(sys.intern, self.features)),
            fix_info=self.fix_info,
        )


DictKey = Union[str, Tuple[str, str]]  # handle or (caller, caller_port)


class DictEntries(TypedDict):
    preconditions: Dict[DictKey, List[ParseConditionTuple]]
    postconditions: Dict[DictKey, List[ParseConditionTuple]]
    issues: Iterable[ParseIssueTuple]


Summary = Dict[str, Any]  # blob of objects that gets passed through the pipeline


# pyre-fixme[3]: Return type must be annotated.
def time_str(delta: timedelta):
    minutes, seconds = divmod(delta.total_seconds(), 60)
    seconds_string = f"{int(seconds)}s"
    if minutes > 0:
        return f"{int(minutes)}m {seconds_string}"
    return seconds_string


class PipelineStep(Generic[T_in, T_out], metaclass=ABCMeta):
    """Pipeline steps have an input type and an output type.
    T_in and T_out should both be child classes of PipelineData.
    """

    # pyre-fixme[3]: Return type must be annotated.
    def __init__(self):
        pass

    @abstractmethod
    def run(self, input: T_in, summary: Summary) -> Tuple[T_out, Summary]:
        raise NotImplementedError("PipelineStep.run is abstract")


class Pipeline(object):
    # pyre-fixme[3]: Return type must be annotated.
    # pyre-fixme[2]: Parameter annotation cannot contain `Any`.
    def __init__(self, steps: List[PipelineStep[Any, Any]]):
        # pyre-fixme[4]: Attribute annotation cannot contain `Any`.
        self.steps: List[PipelineStep[Any, Any]] = steps

    # pyre-fixme[3]: Return annotation cannot contain `Any`.
    def run(
        self,
        # pyre-fixme[2]: Parameter must be annotated.
        first_input,
        summary: Optional[Summary] = None,
    ) -> Tuple[Any, Summary]:
        if summary is None:
            summary = {}
        next_input = first_input
        timing = []
        for step in self.steps:
            start_time = datetime.now()
            next_input, summary = step.run(next_input, summary)
            timing.append((step.__class__.__name__, datetime.now() - start_time))
        log.info(
            "Step timing: %s",
            ", ".join([f"{name} took {time_str(delta)}" for name, delta in timing]),
        )
        return next_input, summary



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import builtins
import enum
import itertools
import os
import sys
from collections import defaultdict
from typing import (
    Any,
    Callable,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import click
import IPython
from IPython import paths
from IPython.core import page
from prompt_toolkit import prompt
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.completion import PathCompleter
from prompt_toolkit.history import FileHistory, History
from pygments import highlight
from pygments.formatters import TerminalFormatter
from pygments.lexers import get_lexer_for_filename
from sqlalchemy.orm import Session, aliased
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.query import Query as RawQuery
from sqlalchemy.orm.util import AliasedClass
from sqlalchemy.sql import func
from sqlalchemy.sql.expression import or_

from .. import queries
from ..analysis_output import AnalysisOutput, AnalysisOutputError
from ..db import DB
from ..decorators import UserError, catch_keyboard_interrupt, catch_user_error
from ..models import (
    DBID,
    Issue,
    IssueInstance,
    IssueInstanceSharedTextAssoc,
    Run,
    RunStatus,
    SharedText,
    SharedTextKind,
    TraceFrame,
    TraceKind,
)
from ..models import create as create_models
from ..pipeline.base_parser import BaseParser
from . import issues as issues_module
from . import trace
from .issues import Instance, IssueQueryResult
from .trace import TraceFrameQueryResult, TraceTuple

T = TypeVar("T")

FilenameText: AliasedClass = aliased(SharedText)
CallableText: AliasedClass = aliased(SharedText)
CallerText: AliasedClass = aliased(SharedText)
CalleeText: AliasedClass = aliased(SharedText)
MessageText: AliasedClass = aliased(SharedText)


class ListFilterException(Exception):
    pass


class LeafOrderBy(str, enum.Enum):
    name = "name"
    number_issues = "number_issues"


ScopeVariables = Dict[str, Union[Callable[..., None], TraceKind]]


class Interactive:
    help_message = """
Commands =======================================================================

== Information commands ==
help                 show this message
help COMMAND         more info about a command
state                show the internal state of the tool for debugging

== Display commands ==
runs                 list all completed static analysis runs
issues               list all issues for the selected run
frames               show trace frames independently of an issue
leaves               list all leaves of issues for the selected run
show                 show info about selected issue or trace frame

== Selection commands ==
analysis_output DIRS sets the location of the analysis output
run ID               select a specific run for browsing issues
latest_run KIND      sets run to the latest of the specified kind
issue ID             select a specific issue for browsing a trace
frame ID             select a trace frame to explore

== Trace commands ==
trace                show a trace of the selected issue or trace frame
prev/p               move backward within the trace
next/n               move forward within the trace
jump NUM             jump to a specific trace frame in a trace
branch INDEX         select a trace branch
list                 show source code at the current trace frame

== Debugging commands ==
parents              show trace frames that call the current trace frame
details              show additional information about the current trace frame
"""
    welcome_message = "Interactive issue exploration. Type 'help' for help."

    SELF_SCOPE_KEY = "_interactive"
    PARSER_CLASS_SCOPE_KEY = "_parser_class"

    def __init__(
        self,
        *,
        database: DB,
        repository_directory: Optional[str] = None,
        parser_class: Type[BaseParser],
        read_only: bool = False,
    ) -> None:
        self.db = database
        self.scope_vars: ScopeVariables = {
            "precondition": TraceKind.PRECONDITION,
            "postcondition": TraceKind.POSTCONDITION,
            "help": self.help,
            "state": self.state,
            "runs": self.runs,
            "issues": self.issues,
            "run": self.run,
            "latest_run": self.latest_run,
            "issue": self.issue,
            "show": self.show,
            "trace": self.trace,
            "next": self.next_cursor_location,
            "n": self.next_cursor_location,
            "prev": self.prev_cursor_location,
            "p": self.prev_cursor_location,
            "jump": self.jump,
            "branch": self.branch,
            "list": self.list_source_code,
            "frames": self.frames,
            "frame": self.frame,
            "leaves": self.leaves,
            "parents": self.parents,
            "details": self.details,
            "analysis_output": self.analysis_output,
            "callable": self.callable,
            self.SELF_SCOPE_KEY: self,
            self.PARSER_CLASS_SCOPE_KEY: parser_class,
        }
        self.repository_directory: str = repository_directory or os.getcwd()
        self.current_analysis_output: Optional[AnalysisOutput] = None

        self._current_run_id: DBID = DBID(-1)

        # Trace exploration relies on either of these
        self.current_issue_instance_id: DBID = DBID(-1)
        self.current_frame_id: DBID = DBID(-1)

        self.sources: Set[str] = set()
        self.sinks: Set[str] = set()
        self.features: Set[str] = set()
        self.sinks_dict: Dict[int, str] = {}

        self._leaf_lookup: trace.LeafLookup = trace.LeafLookup({}, {}, {})

        # Tuples representing the trace of the current issue
        self.trace_tuples: List[TraceTuple] = []
        # Active trace frame of the current trace
        self.current_trace_frame_index: int = -1

        # Persist history for prompts that opt-into it, by specifying
        # history_key on self.prompt().
        self.prompt_history: Dict[str, History] = {}

        self.read_only = read_only

    def setup(self) -> ScopeVariables:
        if not self.read_only:
            create_models(self.db)
        with self.db.make_session() as session:
            latest_run_id = queries.latest_run_id(session)

            self._leaf_lookup = trace.LeafLookup.create(session)

        print("=" * len(self.welcome_message))
        print(self.welcome_message)
        print("=" * len(self.welcome_message))

        if latest_run_id.resolved() is None:
            self.warning(
                "No runs found. "
                f"Try running '{os.path.basename(sys.argv[0])} analyze' first."
            )
        else:
            self._current_run_id = latest_run_id

        return self.scope_vars

    def help(self, object: Optional[object] = None) -> None:
        if object is None:
            print(self.help_message)
            return

        builtins.help(object)

    def state(self) -> None:
        print(f"              Database: {self.db.dbtype}:{self.db.dbname}")
        print(f"       Analysis Output: {self.current_analysis_output}")
        print(f"  Repository directory: {self.repository_directory}")
        print(f"           Current run: {self._current_run_id}")
        print(f"Current issue instance: {int(self.current_issue_instance_id)}")
        print(f"   Current trace frame: {int(self.current_frame_id)}")
        print(f"        Sources filter: {self.sources}")
        print(f"          Sinks filter: {self.sinks}")

    @catch_keyboard_interrupt()
    def runs(self, use_pager: bool = False) -> None:
        pager = self._resolve_pager(use_pager)

        with self.db.make_session() as session:
            runs = session.query(Run).filter(Run.status == RunStatus.FINISHED)

            run_strings = [
                "\n".join([f"Run {run.id}", f"Date: {run.date}", "-" * 80])
                for run in runs
            ]
        run_output = "\n".join(run_strings)

        pager(run_output)
        print(f"Found {len(run_strings)} runs.")

    @catch_keyboard_interrupt()
    def run(self, run_id: DBID) -> None:
        with self.db.make_session() as session:
            selected_run = (
                session.query(Run)
                .filter(Run.status == RunStatus.FINISHED)
                .filter(Run.id == run_id)
                .scalar()
            )

        if selected_run is None:
            self.warning(
                f"Run {run_id} doesn't exist or is not finished. "
                "Type 'runs' for available runs."
            )
            return

        self._current_run_id = selected_run.id
        print(f"Set run to {run_id}.")

    def _get_profile_basedir(self) -> str:
        profile_name = IPython.get_ipython().profile
        return paths.locate_profile(profile=profile_name)

    def _get_prompt_history(self, key: Optional[str]) -> Optional[History]:
        if key is None:
            return None

        history = self.prompt_history.get(key)
        if not history:
            basedir = os.path.join(self._get_profile_basedir(), "prompt_history")
            os.makedirs(basedir, exist_ok=True)
            history = FileHistory(os.path.join(basedir, key))
            self.prompt_history[key] = history
        return history

    def prompt(
        self,
        message: Optional[str] = "",
        history_key: Optional[str] = None,
        *args: Any,
        **kwargs: Any,
    ) -> str:
        try:
            ret = prompt(
                message,
                history=self._get_prompt_history(history_key),
                auto_suggest=AutoSuggestFromHistory(),
                *args,
                **kwargs,
            )
            if ret == "":
                raise KeyboardInterrupt()
            return ret
        except EOFError:
            raise KeyboardInterrupt()

    @catch_keyboard_interrupt()
    @catch_user_error()
    def analysis_output(self, locations: Optional[List[str]] = None) -> None:
        """Sets the location of the output from the static analysis tool.

        Parameters:
            locations: List[str]   Filesystem locations for the results.
        """
        try:
            if not locations:
                locations = [
                    self.prompt(
                        "Analysis results: ",
                        history_key="analysis_results",
                        completer=PathCompleter(),
                    )
                ]
            self.current_analysis_output = AnalysisOutput.from_strs(locations)
        except AnalysisOutputError as e:
            raise UserError(f"Error loading results: {e}")

    @catch_user_error()
    def latest_run(self, run_kind: str) -> None:
        """Sets the current run to the latest run of a given kind.

        Parameters (required):
            run_kind: str    the run kind to filter by

        Example:
            'latest_run "master"' will set the current run to the latest
            run whose kind field is "master"
        """
        if not run_kind or not isinstance(run_kind, str):
            raise UserError("Please provide a non-empty string for 'run_kind'.")

        with self.db.make_session() as session:
            selected_run_id = (
                session.query(func.max(Run.id))
                .filter(Run.kind == run_kind)
                .filter(Run.status == RunStatus.FINISHED)
                .scalar()
            )

        if selected_run_id.resolved() is None:
            raise UserError(f"No runs with kind '{run_kind}'.")

        self._current_run_id = selected_run_id
        print(f"Set run to {self._current_run_id}.")

    @catch_keyboard_interrupt()
    def issue(self, issue_instance_id: DBID) -> None:
        """Select an issue.

        Parameters:
            issue_instance_id: int    id of the issue instance to select

        Note: We are selecting issue instances, even though the command is called
        issue.
        """
        with self.db.make_session() as session:
            selected_issue = (
                session.query(IssueInstance)
                .filter(IssueInstance.id == issue_instance_id)
                .scalar()
            )

            if selected_issue is None:
                self.warning(
                    f"Issue {issue_instance_id} doesn't exist. "
                    "Type 'issues' for available issues."
                )
                return

            self.sources = self._get_leaves_issue_instance(
                session, issue_instance_id, SharedTextKind.SOURCE
            )

            self.sinks = self._get_leaves_issue_instance(
                session, issue_instance_id, SharedTextKind.SINK
            )

            self.features = self._get_leaves_issue_instance(
                session, issue_instance_id, SharedTextKind.FEATURE
            )

        self.current_issue_instance_id = selected_issue.id
        self.current_frame_id = DBID(-1)
        self.current_trace_frame_index = 1  # first one after the source

        print(f"Set issue to {issue_instance_id}.")
        if selected_issue.run_id != self._current_run_id:
            self._current_run_id = selected_issue.run_id
            print(f"Set run to {self._current_run_id}.")
        print()

        self._generate_trace_from_issue()
        self.show()

    @catch_keyboard_interrupt()
    @catch_user_error()
    def show(self) -> None:
        """More details about the selected issue or trace frame."""
        self._verify_entrypoint_selected()

        if int(self.current_issue_instance_id) != -1:
            self._show_current_issue_instance()
            return

        self._show_current_trace_frame()

    @catch_keyboard_interrupt()
    @catch_user_error()
    def issues(
        self,
        use_pager: bool = False,
        *,
        codes: Optional[Union[int, List[int]]] = None,
        callables: Optional[Union[str, List[str]]] = None,
        filenames: Optional[Union[str, List[str]]] = None,
        all_features: Optional[Union[str, List[str]]] = None,
        any_features: Optional[Union[str, List[str]]] = None,
        exclude_features: Optional[Union[str, List[str]]] = None,
        exact_trace_length_to_sources: Optional[int] = None,
        exact_trace_length_to_sinks: Optional[int] = None,
        max_trace_length_to_sources: Optional[int] = None,
        max_trace_length_to_sinks: Optional[int] = None,
        statuses: Optional[Union[str, List[str]]] = None,
    ) -> None:
        """Lists issues for the selected run.

        Parameters (all optional):
            use_pager: bool                use a unix style pager for output
            codes: int or list[int]        issue codes to filter on
            callables: str or list[str]    callables to filter on (supports wildcards)
            filenames: str or list[str]    filenames to filter on (supports wildcards)
            statuses: str or list[str]     statues to filter on (supports wildcards)
            all_features: str or list[str] features to filter on
            any_features: str or list[str] features to inclusively filter on
            exclude_features: str or list[str]
                features to exclude issues based upon
            exact_trace_length_to_sources: int
                exact values for min trace length to sources to filter on
            exact_trace_length_to_sinks: int
                exact values for min trace length to sinks to filter on
            max_trace_length_to_sources: int
                maximum value for min trace length to sources to filter on
            max_trace_length_to_sinks: int
                maximum value for min trace length to sinks to filter on

        String filters support LIKE wildcards (%, _) from SQL:
            % matches anything (like .* in regex)
            _ matches 1 character (like . in regex)

        For example:
            callables=[
                "%occurs.anywhere%",
                "%at.end",
                "at.start%",
                "etc.",
            ])
        """
        pager = self._resolve_pager(use_pager)

        with self.db.make_session() as session:
            builder = Instance(session, self._current_run_id)

            if codes is not None:
                if not isinstance(codes, int) and not isinstance(codes, list):
                    raise UserError("'codes' should be int or list of int.")
                if isinstance(codes, int):
                    codes = [codes]
                builder = builder.where_codes_is_any_of(codes)

            if callables is not None:
                if not isinstance(callables, str) and not isinstance(callables, list):
                    raise UserError("'callables' should be str or list of str.")
                if isinstance(callables, str):
                    callables = [callables]
                builder = builder.where_callables_is_any_of(callables)

            if statuses is not None:
                if not isinstance(statuses, str) and not isinstance(statuses, list):
                    raise UserError("'statuses' should be str or list of str.")
                if isinstance(statuses, str):
                    statuses = [statuses]
                builder = builder.where_status_is_any_of(statuses)

            if filenames is not None:
                if not isinstance(filenames, str) and not isinstance(filenames, list):
                    raise UserError("'filenames' should be str or list of str.")
                if isinstance(filenames, str):
                    filenames = [filenames]
                builder = builder.where_path_is_any_of(filenames)

            if (exact_trace_length_to_sources is not None) and (
                max_trace_length_to_sources is not None
            ):
                raise UserError(
                    (
                        "'exact_trace_length_to_sources' and "
                        "'max_trace_length_to_sources' can't be set together"
                    )
                )

            if (exact_trace_length_to_sinks is not None) and (
                max_trace_length_to_sinks is not None
            ):
                raise UserError(
                    (
                        "'exact_trace_length_to_sinks' and "
                        "'max_trace_length_to_sinks' can't be set together"
                    )
                )

            if exact_trace_length_to_sources is not None:
                if not isinstance(exact_trace_length_to_sources, int):
                    raise UserError("'exact_trace_length_to_sources' should be int.")
                builder = builder.where_trace_length_to_sources(
                    exact_trace_length_to_sources, exact_trace_length_to_sources
                )

            if exact_trace_length_to_sinks is not None:
                if not isinstance(exact_trace_length_to_sinks, int):
                    raise UserError("'exact_trace_length_to_sinks' should be int.")
                builder = builder.where_trace_length_to_sinks(
                    exact_trace_length_to_sinks, exact_trace_length_to_sinks
                )

            if max_trace_length_to_sources is not None:
                if not isinstance(max_trace_length_to_sources, int):
                    raise UserError("'max_trace_length_to_sources' should be int.")
                builder = builder.where_trace_length_to_sources(
                    0, max_trace_length_to_sources
                )

            if max_trace_length_to_sinks is not None:
                if not isinstance(max_trace_length_to_sinks, int):
                    raise UserError("'max_trace_length_to_sinks' should be int.")
                builder = builder.where_trace_length_to_sinks(
                    0, max_trace_length_to_sinks
                )

            if all_features is not None:
                if not isinstance(all_features, str) and not isinstance(
                    all_features, list
                ):
                    raise UserError("'all_features' should be str or list of str.")
                if isinstance(all_features, str):
                    all_features = [all_features]
                builder = builder.where_all_features(all_features)

            if any_features is not None:
                if not isinstance(any_features, str) and not isinstance(
                    any_features, list
                ):
                    raise UserError("'any_features' should be str or list of str.")
                if isinstance(any_features, str):
                    any_features = [any_features]
                builder = builder.where_any_features(any_features)

            if exclude_features is not None:
                if not isinstance(exclude_features, str) and not isinstance(
                    exclude_features, list
                ):
                    raise UserError("'any_features' should be str or list of str.")
                if isinstance(exclude_features, str):
                    exclude_features = [exclude_features]
                builder = builder.where_exclude_features(exclude_features)

            issues = builder.get()
            sources_list = [
                issues_module.sources(session, issue.issue_instance_id)
                for issue in issues
            ]
            sinks_list = [
                issues_module.sinks(session, issue.issue_instance_id)
                for issue in issues
            ]
            features_list = [
                issues_module.features(session, issue.issue_instance_id)
                for issue in issues
            ]

        issue_strings = []
        for issue, sources, sinks, features in zip(
            issues, sources_list, sinks_list, features_list
        ):
            issue_strings.append(
                self._create_issue_output_string(issue, sources, sinks, features)
            )

        issue_output = f"\n{'-' * 80}\n".join(issue_strings)
        pager(issue_output)
        print(f"Found {len(issue_strings)} issues with run_id {self._current_run_id}.")

    @catch_user_error()
    def trace(self, features: bool = False) -> None:
        """Show a trace for the selected issue or trace frame.

        The '-->' token points to the currently active trace frame within the
        trace.

        Trace output has 4 columns:
        - branches: the number of siblings a node has (including itself)
          [indicates that the trace branches into multiple paths]
        - callable: the name of the object that was called
        - port/condition: a description of the type of trace frame
          - source: where data originally comes from
          - root: the main callable through which the data propagates
          - sink: where data eventually flows to
        - location: the relative location of the trace frame's source code

        Example output:
             ⎇  [callable]            [port]    [location]
             +2 leaf                  source    module/main.py:26|4|8
         -->    module.main           root      module/helper.py:76|5|10
                module.helper.process root      module/helper.py:76|5|10
             +3 leaf                  sink      module/main.py:74|1|9
        """

        self._verify_entrypoint_selected()
        self._output_trace_tuples(self.trace_tuples, features)

    @catch_keyboard_interrupt()
    @catch_user_error()
    def frames(
        self,
        *,
        callers: Optional[Union[str, List[str]]] = None,
        callees: Optional[Union[str, List[str]]] = None,
        kind: Optional[TraceKind] = None,
        limit: Optional[int] = 10,
    ) -> None:
        """Display trace frames independent of the current issue.

        Parameters (all optional):
            callers: str or list[str]            filter traces by this caller name
            callees: str or list[str]            filter traces by this callee name
            kind: precondition|postcondition    the type of trace frames to show
            limit: int (default: 10)            how many trace frames to display
                                                (specify limit=None for all)

        Sample usage:
            frames callers="module.function", kind=postcondition

        String filters support LIKE wildcards (%, _) from SQL:
            % matches anything (like .* in regex)
            _ matches 1 character (like . in regex)
        """
        with self.db.make_session() as session:
            query = (
                session.query(
                    TraceFrame.id,
                    CallerText.contents.label("caller"),
                    TraceFrame.caller_port,
                    CalleeText.contents.label("callee"),
                    TraceFrame.callee_port,
                )
                .filter(TraceFrame.run_id == self._current_run_id)
                .join(CallerText, CallerText.id == TraceFrame.caller_id)
                .join(CalleeText, CalleeText.id == TraceFrame.callee_id)
            )

            if callers is not None:
                query = self._add_list_or_string_filter_to_query(
                    callers, query, CallerText.contents, "callers"
                )

            if callees is not None:
                query = self._add_list_or_string_filter_to_query(
                    callees, query, CalleeText.contents, "callees"
                )

            if kind is not None:
                if kind not in {TraceKind.PRECONDITION, TraceKind.POSTCONDITION}:
                    raise UserError(
                        "Try 'frames kind=postcondition'"
                        " or 'frames kind=precondition'."
                    )
                query = query.filter(TraceFrame.kind == kind)

            if limit is not None and not isinstance(limit, int):
                raise UserError("'limit' should be an int or None.")

            trace_frames = query.group_by(TraceFrame.id).order_by(
                CallerText.contents, CalleeText.contents
            )

            total_trace_frames = trace_frames.count()
            limit = limit or total_trace_frames

            self._output_trace_frames(
                self._group_trace_frames(trace_frames, limit), limit, total_trace_frames
            )

    @catch_keyboard_interrupt()
    def frame(self, frame_id: int) -> None:
        with self.db.make_session() as session:
            selected_frame = (
                session.query(TraceFrame.id, TraceFrame.kind, TraceFrame.run_id)
                .filter(TraceFrame.id == frame_id)
                .first()
            )

            if selected_frame is None:
                self.warning(
                    f"Trace frame {frame_id} doesn't exist. "
                    "Type 'frames' for available trace frames."
                )
                return

            if selected_frame.kind == TraceKind.POSTCONDITION:
                self.sinks = set()
                self.sources = trace.get_leaves_trace_frame(
                    session, int(selected_frame.id), SharedTextKind.SOURCE
                )

            else:
                self.sinks = trace.get_leaves_trace_frame(
                    session, int(selected_frame.id), SharedTextKind.SINK
                )

                self.sources = set()

        self.current_frame_id = selected_frame.id
        self.current_issue_instance_id = DBID(-1)

        print(f"Set trace frame to {frame_id}.")
        if int(selected_frame.run_id) != self._current_run_id:
            self._current_run_id = selected_frame.run_id
            print(f"Set run to {self._current_run_id}.")
        print()

        self._generate_trace_from_frame()
        self.show()

    @catch_keyboard_interrupt()
    @catch_user_error()
    def leaves(
        self,
        kind: str = "sink",
        use_pager: bool = False,
        limit: Optional[int] = None,
        order_by: Optional[LeafOrderBy] = None,
    ) -> None:
        """Lists leaves of issues for the selected run.

        Parameters (all optional):
            kind: source|sink|feature (default: sink)       the type of leaves to show
            use_pager: bool                                 use a unix style pager for output
            limit: int (default: all)                       how many leaves to display
            order_by: name|number_issues (default: random)  sort by a criteria
        """
        leaves: DefaultDict[str, int] = defaultdict(int)

        with self.db.make_session() as session:
            query = queries.leaves(
                session=session, kind=kind, run_id=self._current_run_id
            )
            for (_, name) in query:
                leaves[name] += 1

        results: Iterable[Tuple[str, int]]
        if order_by == LeafOrderBy.name:
            results = sorted(leaves.items(), key=lambda leaf: leaf[0])
        elif order_by == LeafOrderBy.number_issues:
            results = sorted(leaves.items(), key=lambda leaf: leaf[1], reverse=True)
        elif order_by is None:
            results = leaves.items()
        else:
            raise UserError("Invalid order_by method.")

        if limit is not None:
            results = itertools.islice(results, limit)

        leaves_strings = []
        for name, number_issues in results:
            leaves_strings.append(f"{name} (in {number_issues} issues)")
        pager = self._resolve_pager(use_pager)
        pager("\n".join(leaves_strings))
        print(
            f"Found {len(leaves)} {kind}s in issues with run_id {self._current_run_id}."
        )

    @catch_keyboard_interrupt()
    @catch_user_error()
    def parents(self) -> None:
        self._verify_entrypoint_selected()
        current_trace_tuple = self.trace_tuples[self.current_trace_frame_index]

        # Don't allow calling from the leaf node in a trace. Instead, call
        # parents() from the placeholder of the caller of the leaf node.
        if current_trace_tuple.trace_frame.is_leaf():
            raise UserError("Try running from a non-leaf node.")

        # Backwards trace checks against the given frame's _caller_.
        # If we have A->B as a placeholder, we want parents that call A, since A
        #   is what is displayed in the trace.
        # If we have A->B as a non-placeholder, we want parents that call B,
        #   since B is what is displayed in the trace. So we go to (A->B)'s child
        #   B->C, and look for frames that call B.
        if not current_trace_tuple.placeholder:
            delta_from_child = -1 if self._is_before_root() else 1
            current_trace_tuple = self.trace_tuples[
                self.current_trace_frame_index + delta_from_child
            ]

        with self.db.make_session() as session:
            if current_trace_tuple.trace_frame.kind == TraceKind.POSTCONDITION:
                leaf_kind = self.sources
            elif current_trace_tuple.trace_frame.kind == TraceKind.PRECONDITION:
                leaf_kind = self.sinks
            else:
                assert (
                    current_trace_tuple.trace_frame.kind == TraceKind.POSTCONDITION
                    or current_trace_tuple.trace_frame.kind == TraceKind.PRECONDITION
                )

            parent_trace_frames = trace.next_frames(
                session,
                current_trace_tuple.trace_frame,
                # pyre-fixme[61]: `leaf_kind` may not be initialized here.
                leaf_kind,
                set(),
                self._current_run_id,
                backwards=True,
            )
        if len(parent_trace_frames) == 0:
            print(
                f"No parents calling [{current_trace_tuple.trace_frame.caller} "
                f": {current_trace_tuple.trace_frame.caller_port}]."
            )
            return

        parent_trace_frame = self._select_parent_trace_frame(parent_trace_frames)

        self._update_trace_tuples_new_parent(parent_trace_frame)
        self.trace()

    def _generate_trace_from_issue(self) -> None:
        with self.db.make_session() as session:
            issue = self._get_current_issue(session)
            postcondition_initial_frames = trace.initial_frames(
                session,
                issue.issue_instance_id,
                TraceKind.POSTCONDITION,
            )
            precondition_initial_frames = trace.initial_frames(
                session,
                issue.issue_instance_id,
                TraceKind.PRECONDITION,
            )

            postcondition_navigation = trace.navigate_trace_frames(
                session,
                postcondition_initial_frames,
                self.sources,
                self.sinks,
            )
            precondition_navigation = trace.navigate_trace_frames(
                session,
                precondition_initial_frames,
                self.sources,
                self.sinks,
            )

        self.trace_tuples = (
            self._create_trace_tuples(reversed(postcondition_navigation))
            + [
                TraceTuple(
                    trace_frame=TraceFrameQueryResult(
                        id=DBID(0),
                        caller="",
                        caller_port="",
                        callee=issue.callable,
                        callee_port="root",
                        filename=issue.filename,
                        callee_location=issue.location,
                    )
                )
            ]
            + self._create_trace_tuples(precondition_navigation)
        )
        self.current_trace_frame_index = len(postcondition_navigation)

    def _generate_trace_from_frame(self) -> None:
        with self.db.make_session() as session:
            trace_frame = (
                session.query(
                    TraceFrame.id,
                    TraceFrame.caller_id,
                    CallerText.contents.label("caller"),
                    TraceFrame.caller_port,
                    TraceFrame.callee_id,
                    CalleeText.contents.label("callee"),
                    TraceFrame.callee_port,
                    TraceFrame.callee_location,
                    TraceFrame.kind,
                    TraceFrame.type_interval_lower,
                    TraceFrame.type_interval_upper,
                    TraceFrame.preserves_type_context,
                    FilenameText.contents.label("filename"),
                )
                .filter(TraceFrame.id == self.current_frame_id)
                .join(CallerText, CallerText.id == TraceFrame.caller_id)
                .join(CalleeText, CalleeText.id == TraceFrame.callee_id)
                .join(FilenameText, FilenameText.id == TraceFrame.filename_id)
                .one()
            )
            navigation = trace.navigate_trace_frames(
                session,
                [TraceFrameQueryResult.from_record(trace_frame)],
                self.sources,
                self.sinks,
            )

        first_trace_frame = navigation[0][0]
        placeholder_tuple = [
            TraceTuple(trace_frame=first_trace_frame, placeholder=True)
        ]

        self.trace_tuples = self._create_trace_tuples(navigation)

        if trace_frame.kind == TraceKind.POSTCONDITION:
            self.trace_tuples = self.trace_tuples[::-1] + placeholder_tuple
            self.current_trace_frame_index = len(self.trace_tuples) - 1
            return

        self.trace_tuples = placeholder_tuple + self.trace_tuples
        self.current_trace_frame_index = 0

    @catch_user_error()
    def next_cursor_location(self) -> None:
        """Move cursor to the next trace frame."""
        self._verify_entrypoint_selected()
        self.current_trace_frame_index = min(
            self.current_trace_frame_index + 1, len(self.trace_tuples) - 1
        )
        self.trace()

    @catch_user_error()
    def prev_cursor_location(self) -> None:
        """Move cursor to the previous trace frame."""
        self._verify_entrypoint_selected()
        self.current_trace_frame_index = max(self.current_trace_frame_index - 1, 0)
        self.trace()

    @catch_user_error()
    def jump(self, selected_number: int) -> None:
        """Jump to a specific trace frame in a trace.

        Parameters:
            selected_number: int    the trace frame number from trace output
        """
        self._verify_entrypoint_selected()
        if selected_number < 1 or selected_number > len(self.trace_tuples):
            raise UserError(
                "Trace frame number out of bounds "
                f"(expected 1-{len(self.trace_tuples)} but got {selected_number})."
            )

        self.current_trace_frame_index = selected_number - 1
        self.trace()

    def _get_branch_options(
        self, session: Session
    ) -> Tuple[List[TraceFrameQueryResult], List[str]]:
        current_trace_tuple = self.trace_tuples[self.current_trace_frame_index]
        filter_leaves = (
            self.sources
            if current_trace_tuple.trace_frame.kind == TraceKind.POSTCONDITION
            else self.sinks
        )

        branches = self._get_trace_frame_branches(session)

        leaves_strings = []
        for frame in branches:
            leaves_strings.append(
                ", ".join(
                    [
                        leaf
                        for leaf in trace.get_leaves_trace_frame(
                            session,
                            int(frame.id),
                            trace.trace_kind_to_shared_text_kind(frame.kind),
                        )
                        if leaf in filter_leaves
                    ]
                )
            )

        return branches, leaves_strings

    @catch_keyboard_interrupt()
    @catch_user_error()
    def branch(self, selected_number: Optional[int] = None) -> None:
        """Show and select branches for a branched trace.
        - [*] signifies the current branch that is selected
        - will prompt for branch selection if called with no argument
        - will automatically select a branch if called with an argument

        Parameters (optional):
            selected_number: int    branch number from expand output

        Example output:

        Suppose we have the trace output:
             ⎇  [callable]            [port]    [location]
         --> +2 leaf                  source    module/main.py:26|4|8
                module.main           root      module/helper.py:76|5|10
                module.helper.process root      module/helper.py:76|5|10
             +3 leaf                  sink      module/main.py:74|1|9

        Calling expand will result in the output:
        [*] leaf
                [0 hops: source]
                [module/main.py:26|4|8]
        [1] module.helper.preprocess
                [1 hops: source]
                [module/main.py:21|4|8]
        """
        self._verify_entrypoint_selected()
        self._verify_multiple_branches()

        with self.db.make_session() as session:
            branches, leaves_strings = self._get_branch_options(session)

            if selected_number is None:
                selected_number = self._select_branch_trace_frame(
                    branches, leaves_strings
                )

            if (
                not isinstance(selected_number, int)
                or selected_number < 1
                or selected_number > len(branches)
            ):
                raise UserError(
                    "Branch number invalid "
                    f"(expected 1-{len(branches)} but got {selected_number})."
                )
            new_navigation = trace.navigate_trace_frames(
                session,
                branches,
                self.sources,
                self.sinks,
                selected_number - 1,
            )

        new_trace_tuples = self._create_trace_tuples(new_navigation)

        if self._is_before_root():
            new_trace_tuples.reverse()
            self.trace_tuples = (
                new_trace_tuples
                + self.trace_tuples[self.current_trace_frame_index + 1 :]
            )

            # If length of prefix changes, it will change some indices
            trace_frame_index_delta = (
                len(new_navigation) - self.current_trace_frame_index - 1
            )
            self.current_trace_frame_index += trace_frame_index_delta
        else:
            self.trace_tuples = (
                self.trace_tuples[: self.current_trace_frame_index] + new_trace_tuples
            )

        self.trace()

    @catch_keyboard_interrupt()
    @catch_user_error()
    def list_source_code(self, context: int = 5) -> None:
        """Show source code around the current trace frame location.

        Parameters:
            context: int    number of lines to show above and below trace location
                            (default: 5)
        """
        self._verify_entrypoint_selected()

        current_trace_frame = self.trace_tuples[
            self.current_trace_frame_index
        ].trace_frame

        filename = os.path.join(
            self.repository_directory, current_trace_frame.filename or "<undefined>"
        )
        file_lines: List[str] = []

        try:
            # Use readlines instead of enumerate(file) because mock_open
            # doesn't support __iter__ until python 3.7.1.
            with open(filename, "r") as file:
                file_lines = file.readlines()
        except FileNotFoundError:
            self.warning(f"Couldn't open {filename}.")
            return

        self._output_file_lines(current_trace_frame, file_lines, context)

    def details(
        self, *, limit: Optional[int] = 5, kind: Optional[TraceKind] = None
    ) -> None:
        """Show additional info about the current trace frame.

        Parameters (all optional):
            limit: int    number of related post/pre conditions to show
                          (pass limit=None to show all)
        """
        self._verify_entrypoint_selected()
        if limit is not None and not isinstance(limit, int):
            raise UserError("'limit' should be an int or None.")
        if kind not in {TraceKind.PRECONDITION, TraceKind.POSTCONDITION, None}:
            raise UserError(
                "Try 'details kind=postcondition'" " or 'details kind=precondition'."
            )

        current_trace_tuple = self.trace_tuples[self.current_trace_frame_index]
        current_trace_frame = current_trace_tuple.trace_frame
        callable, _port = self._get_callable_from_trace_tuple(current_trace_tuple)

        print(self._create_trace_frame_output_string(current_trace_frame))
        print(
            f"\nIssues in callable ({callable}):",
            self._num_issues_with_callable(current_trace_frame.callee),
        )

        if kind is None or kind == TraceKind.POSTCONDITION:
            print(f"\nPostconditions with caller ({callable}):")
            self.frames(callers=callable, kind=TraceKind.POSTCONDITION, limit=limit)

        if kind is None or kind == TraceKind.PRECONDITION:
            print(f"\nPreconditions with caller ({callable}):")
            self.frames(callers=callable, kind=TraceKind.PRECONDITION, limit=limit)

    def warning(self, message: str) -> None:
        print(message, file=sys.stderr)

    def _get_trace_frame_branches(
        self, session: Session
    ) -> List[TraceFrameQueryResult]:
        delta_from_parent = 1 if self._is_before_root() else -1
        parent_index = self.current_trace_frame_index + delta_from_parent
        parent_trace_tuple = self.trace_tuples[parent_index]

        if self._is_root_trace_tuple(parent_trace_tuple):
            kind = (
                TraceKind.POSTCONDITION
                if self._is_before_root()
                else TraceKind.PRECONDITION
            )
            return trace.initial_frames(session, self.current_issue_instance_id, kind)

        parent_trace_frame = self.trace_tuples[parent_index].trace_frame

        if parent_trace_frame.kind == TraceKind.POSTCONDITION:
            leaf_kind = self.sources
        elif parent_trace_frame.kind == TraceKind.PRECONDITION:
            leaf_kind = self.sinks
        else:
            assert (
                parent_trace_frame.kind == TraceKind.POSTCONDITION
                or parent_trace_frame.kind == TraceKind.PRECONDITION
            )

        return trace.next_frames(
            session,
            parent_trace_frame,
            # pyre-fixme[61]: `leaf_kind` may not be initialized here.
            leaf_kind,
            set(),
            self._current_run_id,
        )

    def _is_before_root(self) -> bool:
        trace_tuple = self.trace_tuples[self.current_trace_frame_index]
        return trace_tuple.trace_frame.kind == TraceKind.POSTCONDITION

    def _is_root_trace_tuple(self, trace_tuple: TraceTuple) -> bool:
        return trace_tuple.trace_frame.callee_port == "root"

    def _current_branch_index(self, branches: List[TraceFrameQueryResult]) -> int:
        selected_branch_id = int(
            self.trace_tuples[self.current_trace_frame_index].trace_frame.id
        )
        for i, branch in enumerate(branches):
            if selected_branch_id == int(branch.id):
                return i
        return -1

    def _group_trace_frames(
        self, trace_frames: Iterable[TraceFrameQueryResult], limit: int
    ) -> Dict[Tuple[str, str], List[TraceFrameQueryResult]]:
        """Buckets together trace frames that have the same caller:caller_port."""
        caller_buckets: DefaultDict[
            Tuple[str, str], List[TraceFrameQueryResult]
        ] = defaultdict(list)
        for trace_frame in itertools.islice(trace_frames, limit):
            caller_buckets[(trace_frame.caller, trace_frame.caller_port)].append(
                trace_frame
            )
        return caller_buckets

    def _add_list_or_string_filter_to_query(
        self,
        filter: Union[str, List[str]],
        query: RawQuery[T],
        column: InstrumentedAttribute,
        argument_name: str,
    ) -> RawQuery[T]:
        # pyre-fixme[7]: Expected `RawQuery[Variable[T]]` but got `RawQuery[str]`.
        return self._add_list_or_element_filter_to_query(
            filter,
            # pyre-fixme[6]: Expected `RawQuery[Variable[T]]` for 2nd param but got
            #  `RawQuery[Variable[T]]`.
            query,
            column,
            argument_name,
            str,
        )

    def _add_list_or_int_filter_to_query(
        self,
        filter: Union[int, List[int]],
        query: RawQuery[T],
        column: InstrumentedAttribute,
        argument_name: str,
    ) -> RawQuery[T]:
        # pyre-fixme[7]: Expected `RawQuery[Variable[T]]` but got `RawQuery[int]`.
        return self._add_list_or_element_filter_to_query(
            filter,
            # pyre-fixme[6]: Expected `RawQuery[Variable[T]]` for 2nd param but got
            #  `RawQuery[Variable[T]]`.
            query,
            column,
            argument_name,
            int,
        )

    def _add_list_or_element_filter_to_query(
        self,
        filter: Union[T, List[T]],
        query: RawQuery[T],
        column: InstrumentedAttribute,
        argument_name: str,
        element_type: Type[Union[str, int]],
    ) -> RawQuery[T]:
        if isinstance(filter, element_type):
            return query.filter(column.like(filter))
        if isinstance(filter, list):
            if not filter:
                raise UserError(f"'{argument_name}' should be non-empty.")
            # pyre-fixme[6]: Expected `str` for 1st param but got `T`.
            return query.filter(or_(*[column.like(item) for item in filter]))
        raise UserError(
            f"'{argument_name}' should be {element_type} or " f"list of {element_type}."
        )

    def _add_max_int_filter_to_query(
        self,
        filter: int,
        query: RawQuery[T],
        column: InstrumentedAttribute,
        argument_name: str,
    ) -> RawQuery[T]:
        if isinstance(filter, int):
            return query.filter(column <= filter)
        raise UserError(f"'{argument_name}' should be int.")

    def _output_file_lines(
        self, trace_frame: TraceFrameQueryResult, file_lines: List[str], context: int
    ) -> None:
        print(
            f"In {trace_frame.caller or trace_frame.callee} "
            f"[{trace_frame.filename}:{trace_frame.callee_location}]"
        )
        assert trace_frame.callee_location is not None
        location = trace_frame.callee_location
        center_line_number = location.line_no
        begin_lineno = max(center_line_number - context, 1)
        end_lineno = min(center_line_number + context, len(file_lines))
        lineno_width = len(str(end_lineno))

        lines_to_show = file_lines[begin_lineno - 1 : end_lineno]

        # In both cases this removes trailing newlines on each line.
        if sys.stdout.isatty():
            lines_to_show = highlight(
                "".join(lines_to_show),
                # startinline is to skip the <? check for php lexer
                get_lexer_for_filename(trace_frame.filename, startinline=True),
                TerminalFormatter(),
            ).split("\n")
        else:
            lines_to_show = [line.rstrip("\n") for line in lines_to_show]

        for i in range(begin_lineno, end_lineno + 1):
            line = lines_to_show[i - begin_lineno]

            prefix = " --> " if i == center_line_number else " " * 5
            prefix += f"{i:<{lineno_width}} "
            print(f"{prefix} {line}")
            if i == center_line_number:
                print(
                    " " * (len(prefix) + location.begin_column),
                    "^" * (location.end_column - location.begin_column),
                )

    def _output_trace_expansion(
        self, trace_frames: List[TraceFrameQueryResult], leaves_strings: List[str]
    ) -> None:
        for i, (frame, leaves) in enumerate(zip(trace_frames, leaves_strings)):
            prefix = (
                "[*]" if i == self._current_branch_index(trace_frames) else f"[{i + 1}]"
            )
            print(f"{prefix} {frame.callee} : {frame.callee_port}")
            print(f"{' ' * 8}[{frame.trace_length} hops: {leaves}]")
            print(f"{' ' * 8}[{frame.filename}:{frame.callee_location}]")

    def _output_trace_frames(
        self,
        trace_buckets: Dict[Tuple[str, str], List[TraceFrameQueryResult]],
        limit: int,
        total_trace_frames: int,
    ) -> None:
        if not trace_buckets:
            print("No trace frames found.")
            return

        max_len_id = max(
            max(
                max(len(str(trace_frame.id)) for trace_frame in value)
                for key, value in trace_buckets.items()
            ),
            len("[id]"),
        )

        print(f"{'[id]':{max_len_id}} [caller:caller_port -> callee:callee_port]")

        callers = trace_buckets.keys()
        for caller, caller_port in callers:
            print(f"{'-' * max_len_id} {caller}:{caller_port} ->")
            trace_frames = trace_buckets[(caller, caller_port)]
            for trace_frame in trace_frames:
                print(
                    f"{int(trace_frame.id):<{max_len_id}} "
                    f"    {trace_frame.callee}:{trace_frame.callee_port}"
                )

        if limit < total_trace_frames:
            print(
                f"...\nShowing {limit}/{total_trace_frames} matching frames. "
                "To see more, call 'frames' with the 'limit' argument."
            )

    def _output_trace_tuples(
        self, trace_tuples: Sequence[TraceTuple], features: bool = False
    ) -> None:
        expand = "+"
        max_length_index = len(str(len(trace_tuples) - 1)) + 1
        max_length_split = max(
            max(
                len(str(trace_tuple.branches)) + len(expand)
                for trace_tuple in trace_tuples
            ),
            len("⎇"),
        )
        max_length_callable = max(
            max(
                len(self._get_callable_from_trace_tuple(trace_tuple)[0])
                for trace_tuple in trace_tuples
            ),
            len("[callable]"),
        )
        max_length_condition = max(
            max(
                len(self._get_callable_from_trace_tuple(trace_tuple)[1])
                for trace_tuple in trace_tuples
            ),
            len("[port]"),
        )

        print(  # table header
            f"{' ' * 5}"
            f"{'#':{max_length_index}}"
            f"{'⎇':{max_length_split}}"
            f" {'[callable]':{max_length_callable}}"
            f" {'[port]':{max_length_condition}}"
            f" [location]"
        )

        for i, trace_tuple in enumerate(trace_tuples):
            prefix = "-->" if i == self.current_trace_frame_index else " " * 3
            prefix += f" {(i + 1):<{max_length_index}}"

            if trace_tuple.missing:
                print(
                    f" {prefix}"
                    f" [Missing trace frame: {trace_tuple.trace_frame.callee}:"
                    f"{trace_tuple.trace_frame.callee_port}]"
                )
                continue

            callable, callable_port = self._get_callable_from_trace_tuple(trace_tuple)
            branches_string = (
                f"{expand}"
                f"{str(trace_tuple.branches):{max_length_split - len(expand)}}"
                if trace_tuple.branches > 1
                else " " * max_length_split
            )
            output_string = (
                f" {prefix}"
                f"{branches_string}"
                f" {callable:{max_length_callable}}"
                f" {callable_port:{max_length_condition}}"
                f" {trace_tuple.trace_frame.filename}"
                f":{trace_tuple.trace_frame.callee_location}"
            )

            print(output_string)
            if features:
                frame_features = [
                    text.contents
                    for text in trace_tuple.trace_frame.shared_texts
                    if text.kind is SharedTextKind.FEATURE
                ]
                if frame_features:
                    prefix = " " * 11 + "--F:"
                    print(f"{prefix} {frame_features}")

    def _create_trace_tuples(
        self, navigation: Iterable[Tuple[TraceFrameQueryResult, int]]
    ) -> List[TraceTuple]:
        return [
            TraceTuple(
                trace_frame=trace_frame,
                branches=branches,
                missing=trace_frame.caller == "",
            )
            for trace_frame, branches in navigation
        ]

    def _create_issue_output_string(
        self,
        issue: IssueQueryResult,
        sources: Set[str],
        sinks: Set[str],
        features: Set[str],
    ) -> str:
        sources_output = f"\n{' ' * 18}".join(sources)
        sinks_output = f"\n{' ' * 18}".join(sinks)
        features_output = f"\n{' ' * 18}".join(features)
        return "\n".join(
            [
                f"Issue {issue.issue_instance_id}",
                f"            Code: {issue.code}",
                f"         Message: {issue.message}",
                f"        Callable: {issue.callable}",
                (
                    f"         Sources: "
                    f"{sources_output if sources_output else 'No sources'}"
                ),
                (
                    f"           Sinks: "
                    f"{sinks_output if sinks_output else 'No sinks'}"
                ),
                (
                    f"        Features: "
                    f"{features_output if features_output else 'No features'}"
                ),
                (f"        Location: " f"{issue.filename}" f":{issue.location}"),
                (
                    f"Min Trace Length: "
                    f"Source ({issue.min_trace_length_to_sources}) | "
                    f"Sink ({issue.min_trace_length_to_sinks})"
                ),
            ]
        )

    def _create_trace_frame_output_string(
        self, trace_frame: TraceFrameQueryResult
    ) -> str:
        leaf_kind = trace.trace_kind_to_shared_text_kind(trace_frame.kind)
        leaves_label = "Sources" if leaf_kind == SharedTextKind.SOURCE else "Sinks"

        with self.db.make_session() as session:
            leaves_output = f"\n{' ' * 13}".join(
                trace.get_leaves_trace_frame(session, trace_frame.id, leaf_kind)
            )

        return "\n".join(
            [
                f"Trace frame {trace_frame.id}",
                f"     Caller: {trace_frame.caller} : {trace_frame.caller_port}",
                f"     Callee: {trace_frame.callee} : {trace_frame.callee_port}",
                f"       Kind: {trace_frame.kind}",
                f"{leaves_label:>{11}}: {leaves_output}",
                (
                    f"   Location: {trace_frame.filename}"
                    f":{trace_frame.callee_location}"
                ),
            ]
        )

    def _select_parent_trace_frame(
        self, parent_trace_frames: List[TraceFrameQueryResult]
    ) -> TraceFrameQueryResult:
        for i, parent in enumerate(parent_trace_frames):
            print(f"[{i + 1}] {parent.caller} : {parent.caller_port}")
        parent_number = self._prompt_for_number(
            "Parent number", len(parent_trace_frames)
        )
        return parent_trace_frames[parent_number - 1]

    def _select_branch_trace_frame(
        self, branch_trace_frames: List[TraceFrameQueryResult], leaves_string: List[str]
    ) -> int:
        self._output_trace_expansion(branch_trace_frames, leaves_string)
        return self._prompt_for_number("Branch number", len(branch_trace_frames))

    def _prompt_for_number(self, prefix: str, max_valid: int) -> int:
        try:
            selected_number = click.prompt(
                f"{prefix} (1-{max_valid}, ctrl-D to abort)", type=int
            )
            if selected_number < 1 or selected_number > max_valid:
                raise UserError("Out of bounds.")
            print()
            return selected_number
        except click.Abort:
            raise KeyboardInterrupt()

    def _update_trace_tuples_new_parent(
        self, parent_trace_frame: TraceFrameQueryResult
    ) -> None:
        # Construct the placeholder trace tuple and the actual one.
        new_head = [
            TraceTuple(trace_frame=parent_trace_frame, placeholder=True),
            TraceTuple(trace_frame=parent_trace_frame),
        ]

        if parent_trace_frame.kind == TraceKind.POSTCONDITION:
            # If current state is: C in [A,B,C,D,E]
            # Then new state is:   [A,B] + new_tail
            new_tail = new_head[::-1]
            self.trace_tuples = (
                self.trace_tuples[: self.current_trace_frame_index] + new_tail
            )
            self.current_trace_frame_index = len(self.trace_tuples) - 1
            return

        # If current state is: C in [A,B,C,D,E]
        # Then new state is:   new_head + [D,E]
        self.trace_tuples = (
            new_head + self.trace_tuples[self.current_trace_frame_index + 1 :]
        )
        self.current_trace_frame_index = 0

    # pyre-fixme[3]: Return type must be annotated.
    def _resolve_pager(self, use_pager: bool):
        use_pager = sys.stdin.isatty() if use_pager is None else use_pager
        return page.page if use_pager else page.display_page

    def _get_current_issue(self, session: Session) -> IssueQueryResult:
        return (
            session.query(
                # pyre-ignore[16]: SQLAlchemy
                IssueInstance.id.label("issue_instance_id"),
                FilenameText.contents.label("filename"),
                IssueInstance.location,
                Issue.code,
                CallableText.contents.label("callable"),
                MessageText.contents.label("message"),
                IssueInstance.min_trace_length_to_sources,
                IssueInstance.min_trace_length_to_sinks,
            )
            .filter(IssueInstance.id == self.current_issue_instance_id)
            .join(Issue, IssueInstance.issue_id == Issue.id)
            .join(FilenameText, FilenameText.id == IssueInstance.filename_id)
            .join(CallableText, CallableText.id == IssueInstance.callable_id)
            .join(MessageText, MessageText.id == IssueInstance.message_id)
            .first()
        )

    def _get_leaves_issue_instance(
        self, session: Session, issue_instance_id: DBID, kind: SharedTextKind
    ) -> Set[str]:
        ids = [
            int(id)
            for id, in session.query(SharedText.id)
            .distinct(SharedText.id)
            .join(
                IssueInstanceSharedTextAssoc,
                SharedText.id == IssueInstanceSharedTextAssoc.shared_text_id,
            )
            .filter(IssueInstanceSharedTextAssoc.issue_instance_id == issue_instance_id)
            .filter(SharedText.kind == kind)
        ]
        return self._leaf_lookup.resolve(ids, kind)

    def _show_current_issue_instance(self) -> None:
        with self.db.make_session() as session:
            issue = self._get_current_issue(session)

        page.display_page(
            self._create_issue_output_string(
                issue, self.sources, self.sinks, self.features
            )
        )

    def _show_current_trace_frame(self) -> None:
        with self.db.make_session() as session:
            trace_frame = (
                session.query(
                    TraceFrame.id,
                    CallerText.contents.label("caller"),
                    TraceFrame.caller_port,
                    CalleeText.contents.label("callee"),
                    TraceFrame.callee_port,
                    TraceFrame.kind,
                    FilenameText.contents.label("filename"),
                    TraceFrame.callee_location,
                )
                .filter(TraceFrame.id == self.current_frame_id)
                .join(CallerText, CallerText.id == TraceFrame.caller_id)
                .join(CalleeText, CalleeText.id == TraceFrame.callee_id)
                .join(FilenameText, FilenameText.id == TraceFrame.filename_id)
                .one()
            )

        page.display_page(self._create_trace_frame_output_string(trace_frame))

    def callable(self) -> Optional[str]:
        """Show the name of the current callable in the trace"""
        if self.current_trace_frame_index != -1:
            return self._get_callable_from_trace_tuple(
                self.trace_tuples[self.current_trace_frame_index]
            )[0]
        return None

    def _verify_entrypoint_selected(self) -> None:
        assert (
            int(self.current_issue_instance_id) == -1
            or int(self.current_frame_id) == -1
        )

        if (
            int(self.current_issue_instance_id) == -1
            and int(self.current_frame_id) == -1
        ):
            raise UserError(
                "Use 'issue ID' or 'frame ID' to select an entrypoint first."
            )

    def _verify_multiple_branches(self) -> None:
        current_trace_tuple = self.trace_tuples[self.current_trace_frame_index]
        if current_trace_tuple.branches < 2:
            raise UserError("This trace frame has no alternate branches to take.")

    def _num_issues_with_callable(self, callable: str) -> int:
        with self.db.make_session() as session:
            return (
                session.query(func.count(IssueInstance.id))
                .join(CallableText, CallableText.id == IssueInstance.callable_id)
                .filter(CallableText.contents == callable)
                .scalar()
            )

    def _get_callable_from_trace_tuple(
        self, trace_tuple: TraceTuple
    ) -> Tuple[str, str]:
        """Returns either (caller, caller_port) or (callee, callee_port)."""
        trace_frame = trace_tuple.trace_frame
        if trace_tuple.placeholder:
            return trace_frame.caller, trace_frame.caller_port
        return trace_frame.callee, trace_frame.callee_port



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import List, NamedTuple, Optional, Type

import click

from .db import DB
from .pipeline.base_parser import BaseParser


class Context(NamedTuple):
    database: DB
    parser_class: Type[BaseParser]
    repository: Optional[str]
    ipython_extensions: List[str] = []
    tool: str = ""


# pyre-fixme[5]: Global expression must be annotated.
pass_context = click.make_pass_decorator(Context, ensure=True)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Generator
from unittest import TestCase, mock

from ..decorators import (
    UserError,
    catch_keyboard_interrupt,
    catch_user_error,
    log_time,
    retryable,
)


class RetryableTest(TestCase):
    def setUp(self) -> None:
        self.times_through = 0

    @retryable(num_tries=5)
    def semiRaiseException(self) -> bool:
        self.times_through += 1
        if self.times_through < 3:
            raise Exception("You are killing me.")
        else:
            return True

    @retryable(num_tries=5, retryable_exs=[ImportError])
    # pyre-fixme[3]: Return type must be annotated.
    def raiseRetryableException(self):
        self.times_through += 1
        if self.times_through < 3:
            raise ImportError
        else:
            raise Exception("You are killing me.")

    def testRetries(self) -> None:
        self.assertTrue(self.semiRaiseException())
        self.assertTrue(
            self.times_through == 3, "times_through = %d" % (self.times_through)
        )
        self.assertTrue(self.semiRaiseException())
        self.assertTrue(
            self.times_through == 4, "times_through = %d" % (self.times_through)
        )

    def testRetryableExceptions(self) -> None:
        self.assertRaises(Exception, lambda: self.raiseRetryableException())
        self.assertEqual(3, self.times_through)


def mocked_time_generator() -> Generator[float, None, None]:
    """
    Returns time in 10s increments
    """
    start = 578854800.0
    while True:
        yield start
        start += 10


@mock.patch("time.time", side_effect=mocked_time_generator())
class LogTimeTest(TestCase):
    @log_time
    def takes_some_time(self) -> None:
        pass

    # pyre-fixme[2]: Parameter must be annotated.
    def testBasic(self, mocked_time_generator) -> None:
        with self.assertLogs("sapp") as context_manager:
            self.takes_some_time()
        self.assertEqual(
            context_manager.output,
            [
                "INFO:sapp:Takes_Some_Time starting...",
                "INFO:sapp:Takes_Some_Time finished (0:00:20)",
            ],
        )


class CatchUserErrorTest(TestCase):
    @catch_user_error()
    # pyre-fixme[3]: Return type must be annotated.
    def throwsUserError(self):
        raise UserError

    def testCatchesUserError(self) -> None:
        try:
            self.throwsUserError()
        except UserError:
            self.fail("Unexpected UserError")

    @catch_user_error()
    # pyre-fixme[3]: Return type must be annotated.
    def throwsException(self):
        raise ValueError

    def testDoesNotCatchOtherExceptions(self) -> None:
        with self.assertRaises(ValueError):
            self.throwsException()


class CatchKeyboardInterruptTest(TestCase):
    @catch_keyboard_interrupt()
    # pyre-fixme[3]: Return type must be annotated.
    def throwsKeyboardInterrupt(self):
        raise KeyboardInterrupt

    def testCatchesKeyboardInterrupt(self) -> None:
        try:
            self.throwsKeyboardInterrupt()
        except KeyboardInterrupt:
            self.fail("Unexpected KeyboardInterrupt")

    @catch_keyboard_interrupt()
    # pyre-fixme[3]: Return type must be annotated.
    def throwsException(self):
        raise ValueError

    def testDoesNotCatchOtherExceptions(self) -> None:
        with self.assertRaises(ValueError):
            self.throwsException()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import sys
from enum import Enum
from typing import Dict, Union, List


if sys.version_info >= (3, 10):
    from typing import TypeAlias
else:
    from typing_extensions import TypeAlias


class SARIFSeverityLevel(Enum):
    WARNING = "warning"
    ERROR = "error"
    NOTE = "note"
    NONE = "none"

    def __str__(self) -> str:
        return self.value


SARIFRegionObject: TypeAlias = Dict[str, int]

SARIFResult: TypeAlias = Dict[
    str,
    Union[
        Dict[str, str],
        List[Dict[str, Dict[str, Union[SARIFRegionObject, Dict[str, str]]]]],
        str,
    ],
]



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from collections import deque
from typing import Set, Tuple

from ..models import DBID, SharedText, SharedTextKind
from ..trace_graph import LeafMapping, TraceGraph
from . import PipelineStep, Summary

# pyre-fixme[5]: Global expression must be annotated.
log = logging.getLogger("sapp")


class AddIssueInstanceLeaves(PipelineStep[TraceGraph, TraceGraph]):
    """For all issues with a given code, adds a leaf at the issue frame and adds
    the relevant trace_frame_leaf_assocs. The depth increases as the trace
    frames get further away (in the caller->callee direction) from the issue
    frame. Useful for queries in the callee->caller direction."""

    def __init__(self, code: int, leaf_name: str, leaf_kind: SharedTextKind) -> None:
        super().__init__()
        self.code = code
        self.leaf_name = leaf_name
        self.leaf_kind = leaf_kind

    def run(self, input: TraceGraph, summary: Summary) -> Tuple[TraceGraph, Summary]:
        graph = input

        # Get all the issue instances within this category
        instances = [
            instance
            for instance in graph.get_issue_instances()
            if self.code == graph.get_issue(instance.issue_id).code
        ]

        # Add all the trace frame ids for all the issue instances
        trace_frames = []
        for instance in instances:
            trace_frames.extend(graph.get_issue_instance_trace_frames(instance))

        # Explore forward (caller -> callee; issue -> leaf)
        queue = deque([(frame, 0) for frame in trace_frames])
        depth_by_frame_id = {}
        while len(queue) > 0:
            trace_frame, depth = queue.popleft()
            trace_frame_id = trace_frame.id.local_id

            # Skip repeat frames unless we arrived at them by a shorter path.
            if (
                trace_frame_id in depth_by_frame_id
                and depth >= depth_by_frame_id[trace_frame_id]
            ):
                continue
            else:
                # Record the minimum depth.
                depth_by_frame_id[trace_frame_id] = depth

            # Iterate forward
            queue.extend(
                (next_frame, depth + 1)
                for next_frame in graph.get_next_trace_frames(trace_frame)
            )

        # Create new leaves based on these depths
        leaf = graph.get_shared_text(self.leaf_kind, self.leaf_name)
        if leaf is None:
            leaf = SharedText.Record(
                id=DBID(), contents=self.leaf_name, kind=self.leaf_kind
            )
            graph.add_shared_text(leaf)

        # Add the assoc to the leaf
        log.info(
            'Adding %d "%s" leaves from issues with code %d...',
            len(depth_by_frame_id),
            self.leaf_name,
            self.code,
        )
        for trace_frame_id, depth in depth_by_frame_id.items():
            trace_frame = graph.get_trace_frame_from_id(trace_frame_id)
            leaf_mapping: Set[LeafMapping] = trace_frame.leaf_mapping
            if leaf.kind == SharedTextKind.source or leaf.kind == SharedTextKind.sink:
                leaf_mapping.add(
                    LeafMapping(leaf.id.local_id, leaf.id.local_id, leaf.id.local_id)
                )
            graph.add_trace_frame_leaf_assoc(trace_frame, leaf, depth)

        return graph, summary



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

from typing import List

import graphene
from sqlalchemy.orm import Session

from ..models import Issue, IssueInstance, SharedText, SharedTextKind


class Code(graphene.ObjectType):
    code = graphene.Int()


def all_codes(session: Session) -> List[Code]:
    return session.query(Issue.code.distinct().label("code")).all()


class Path(graphene.ObjectType):
    path = graphene.String()


def all_paths(session: Session) -> List[Path]:
    return (
        # pyre-fixme[16]: `str` has no attribute `label`.
        session.query(IssueInstance, SharedText.contents.label("path"))
        .join(SharedText, SharedText.id == IssueInstance.filename_id)
        .group_by(SharedText)
        .all()
    )


class SourceName(graphene.ObjectType):
    source_name = graphene.String()


def all_source_names(session: Session) -> List[SourceName]:
    return (
        # pyre-fixme[16]: `str` has no attribute `label`
        session.query(IssueInstance, SharedText.contents.label("source_name"))
        .join(SharedText, SharedText.kind == SharedTextKind.SOURCE_DETAIL)
        .group_by(SharedText)
        .all()
    )


class SourceKind(graphene.ObjectType):
    source_kind = graphene.String()


def all_source_kinds(session: Session) -> List[SourceName]:
    return (
        # pyre-fixme[16]: `str` has no attribute `label`
        session.query(IssueInstance, SharedText.contents.label("source_kind"))
        .join(SharedText, SharedText.kind == SharedTextKind.SOURCE)
        .group_by(SharedText)
        .all()
    )


class SinkName(graphene.ObjectType):
    sink_name = graphene.String()


def all_sink_names(session: Session) -> List[SourceName]:
    return (
        # pyre-fixme[16]: `str` has no attribute `label`
        session.query(IssueInstance, SharedText.contents.label("sink_name"))
        .join(SharedText, SharedText.kind == SharedTextKind.SINK_DETAIL)
        .group_by(SharedText)
        .all()
    )


class SinkKind(graphene.ObjectType):
    sink_kind = graphene.String()


def all_sink_kinds(session: Session) -> List[SourceName]:
    return (
        # pyre-fixme[16]: `str` has no attribute `label`
        session.query(IssueInstance, SharedText.contents.label("sink_kind"))
        .join(SharedText, SharedText.kind == SharedTextKind.SINK)
        .group_by(SharedText)
        .all()
    )


class Status(graphene.ObjectType):
    status = graphene.String()


def all_statuses(session: Session) -> List[Status]:
    return session.query(Issue.status.distinct().label("status")).all()


class Callable(graphene.ObjectType):
    callable = graphene.String()


def all_callables(session: Session) -> List[Callable]:
    return (
        # pyre-fixme[16]: `str` has no attribute `label`.
        session.query(IssueInstance, SharedText.contents.label("callable"))
        .join(SharedText, SharedText.id == IssueInstance.callable_id)
        .group_by(SharedText)
        .all()
    )


class Feature(graphene.ObjectType):
    feature = graphene.String()


def all_features(session: Session) -> List[Feature]:
    return (
        # pyre-fixme[16]: `str` has no attribute `label`.
        session.query(SharedText, SharedText.contents.label("feature"))
        .filter(SharedText.kind == SharedTextKind.FEATURE)
        .all()
    )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import contextlib
import os
import unittest
from typing import Generator
from unittest import TestCase
from unittest.mock import patch

from click.testing import CliRunner

from .. import __name__ as client
from ..cli import cli

PIPELINE_RUN = f"{client}.pipeline.Pipeline.run"


@contextlib.contextmanager
def isolated_fs() -> Generator[str, None, None]:
    with CliRunner().isolated_filesystem() as f:
        os.mkdir(".hg")
        yield f


@patch(
    f"{client}.analysis_output.AnalysisOutput.from_directory",
    return_value="fake_analysis_output",
)
class TestSappCli(TestCase):
    def setUp(self) -> None:
        self.runner = CliRunner()

    @unittest.skip("T41451811")
    # pyre-fixme[2]: Parameter must be annotated.
    def test_explore_options(self, mock_analysis_output) -> None:
        with isolated_fs():
            result = self.runner.invoke(
                cli,
                [
                    "--database-engine",
                    "memory",
                    "--database-name",
                    "mydatabase",
                    "explore",
                ],
            )
            print(result.output)
            self.assertEqual(result.exit_code, 0)

    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def verify_input_file(self, inputfile, summary_blob) -> None:
        self.assertEqual(inputfile, "fake_analysis_output")

    # pyre-fixme[2]: Parameter must be annotated.
    def test_input_file(self, mock_analysis_output) -> None:
        with patch(PIPELINE_RUN, self.verify_input_file):
            with isolated_fs() as path:
                result = self.runner.invoke(
                    cli, ["--database-name", "sapp.db", "analyze", path]
                )
                print(result.output)
                self.assertEqual(result.exit_code, 0)

    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def verify_base_summary_blob(self, input_files, summary_blob) -> None:
        self.assertEqual(summary_blob["run_kind"], "master")
        self.assertEqual(summary_blob["repository"][:4], "/tmp")
        self.assertEqual(summary_blob["branch"], "master")
        self.assertEqual(summary_blob["commit_hash"], "abc123")
        self.assertEqual(summary_blob["old_linemap_file"][:4], "/tmp")
        self.assertEqual(summary_blob["store_unused_models"], True)

    # pyre-fixme[2]: Parameter must be annotated.
    def test_base_summary_blob(self, mock_analysis_output) -> None:
        with patch(PIPELINE_RUN, self.verify_base_summary_blob):
            with isolated_fs() as path:
                result = self.runner.invoke(
                    cli,
                    [
                        "analyze",
                        "--run-kind",
                        "master",
                        "--branch",
                        "master",
                        "--commit-hash",
                        "abc123",
                        "--linemap",
                        path,
                        "--store-unused-models",
                        path,
                    ],
                )
                self.assertEqual(result.exit_code, 0)

    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def verify_option_job_id(self, input_files, summary_blob) -> None:
        self.assertEqual(summary_blob["job_id"], "job-id-1")

    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def verify_option_job_id_none(self, input_files, summary_blob) -> None:
        self.assertIsNone(summary_blob["job_id"])

    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def verify_option_differential_id(self, input_files, summary_blob) -> None:
        self.assertEqual(summary_blob["job_id"], "user_input_1234567")

    # pyre-fixme[2]: Parameter must be annotated.
    def test_option_job_id(self, mock_analysis_output) -> None:
        with patch(PIPELINE_RUN, self.verify_option_job_id):
            with isolated_fs() as path:
                result = self.runner.invoke(
                    cli, ["analyze", "--job-id", "job-id-1", path]
                )
                self.assertEqual(result.exit_code, 0)

        with patch(PIPELINE_RUN, self.verify_option_job_id_none):
            with isolated_fs() as path:
                result = self.runner.invoke(cli, ["analyze", path])
                print(result.stdout)
                self.assertEqual(result.exit_code, 0)

        with patch(PIPELINE_RUN, self.verify_option_differential_id):
            with isolated_fs() as path:
                result = self.runner.invoke(
                    cli, ["analyze", "--differential-id", "1234567", path]
                )
                self.assertEqual(result.exit_code, 0)

    # pyre-fixme[2]: Parameter must be annotated.
    # pyre-fixme[2]: Parameter must be annotated.
    def verify_previous_issue_handles(self, input_files, summary_blob) -> None:
        self.assertEqual(summary_blob["previous_issue_handles"], "fake_analysis_output")

    @patch(
        f"{client}.analysis_output.AnalysisOutput.from_file",
        return_value="fake_analysis_output",
    )
    # pyre-fixme[2]: Parameter must be annotated.
    def test_previous_input(self, _, mock_analysis_output) -> None:
        with patch(PIPELINE_RUN, self.verify_previous_issue_handles):
            with isolated_fs() as path:
                result = self.runner.invoke(
                    cli,
                    [
                        "analyze",
                        "--previous-issue-handles",
                        path,
                        path,
                    ],
                )
            self.assertEqual(result.exit_code, 0)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Tuple

from ..trace_graph import TraceGraph
from ..trimmed_trace_graph import TrimmedTraceGraph
from . import PipelineStep, Summary

log: logging.Logger = logging.getLogger("sapp")


class TrimTraceGraph(PipelineStep[TraceGraph, TraceGraph]):
    def run(self, input: TraceGraph, summary: Summary) -> Tuple[TraceGraph, Summary]:
        if summary.get("affected_files") is None:
            return input, summary

        log.info("Trimming graph to affected files.")
        trimmed_graph = TrimmedTraceGraph(
            summary["affected_files"], summary.get("affected_issues_only", False)
        )
        trimmed_graph.populate_from_trace_graph(input)
        return trimmed_graph, summary



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import json
import logging
from pathlib import Path

from sqlalchemy import Table
from sqlalchemy.engine import Connection

from . import models
from .db import DB
from .models import WarningMessage

log: logging.Logger = logging.getLogger()


def upsert_entry(
    database_connection: Connection,
    warning_messages_table: Table,
    code: int,
    message: str,
) -> None:
    warning = database_connection.execute(
        warning_messages_table.select().where(warning_messages_table.c.code == code)
    ).first()

    if warning and warning.message != message:
        database_connection.execute(
            warning_messages_table.update()
            .where(warning_messages_table.c.code == code)
            .values(message=message)
        )
        log.info(f"Updated - {code}: {message}")
    elif not warning:
        database_connection.execute(
            warning_messages_table.insert().values(code=code, message=message)
        )
        log.info(f"Added - {code}: {message}")


def update_warning_messages(database: DB, metadata_file: Path) -> None:
    warning_messages_from_metadata_file = json.loads(metadata_file.read_text())["codes"]
    warning_messages = {
        int(code): message
        for code, message in warning_messages_from_metadata_file.items()
    }

    # TODO(T89343050)
    models.create(database)

    warning_messages_table = Table(
        WarningMessage.__tablename__, WarningMessage.metadata
    )

    with database.make_session() as session:
        with session.connection() as database_connection:
            for code, message in warning_messages.items():
                upsert_entry(database_connection, warning_messages_table, code, message)
            session.commit()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
from unittest import TestCase

from ..filter import FilterValidationException, StoredFilter

filter_dictionary = {
    "name": "test filter",
    "description": "this is a description",
    "features": [
        {"mode": "all of", "features": ["first-index:name", "has:first-field"]},
        {"mode": "any of", "features": ["always-via:format-string"]},
        {
            "mode": "none of",
            "features": ["always-via:format-string", "first-index:<unknown>"],
        },
    ],
    "codes": [5005],
    "paths": ["main.py"],
    "callables": ["main.lookup", "main.listservices", "main.evaluate"],
    "traceLengthFromSources": [0, 16],
    "traceLengthToSinks": [0, 19],
    "is_new_issue": True,
}


class StoredFilterTests(TestCase):
    def _test_filter_equality(
        self,
        storedfilter_instance: StoredFilter,
        test_name: bool = True,
        test_description: bool = True,
        test_features: bool = True,
        test_codes: bool = True,
        test_paths: bool = True,
        test_callables: bool = True,
        test_traceLengthFromSources: bool = True,
        test_traceLengthToSinks: bool = True,
        test_is_new_issue: bool = True,
    ) -> None:
        if test_name:
            self.assertEqual(filter_dictionary.get("name"), storedfilter_instance.name)
            self.assertIsInstance(storedfilter_instance.name, str)

        if test_description:
            self.assertEqual(
                filter_dictionary.get("description"), storedfilter_instance.description
            )
            self.assertIsInstance(storedfilter_instance.description, str)

        if test_features:
            self.assertEqual(
                filter_dictionary.get("features"),
                storedfilter_instance.features,
            )
            self.assertIsInstance(
                storedfilter_instance.features,
                list,
            )

        if test_codes:
            self.assertEqual(
                filter_dictionary.get("codes"),
                storedfilter_instance.codes,
            )
            self.assertIsInstance(
                storedfilter_instance.codes,
                list,
            )

        if test_paths:
            self.assertEqual(
                filter_dictionary.get("paths"),
                storedfilter_instance.paths,
            )
            self.assertIsInstance(
                storedfilter_instance.paths,
                list,
            )

        if test_callables:
            self.assertEqual(
                filter_dictionary.get("callables"),
                storedfilter_instance.callables,
            )
            self.assertIsInstance(
                storedfilter_instance.callables,
                list,
            )

        if test_traceLengthFromSources:
            self.assertEqual(
                filter_dictionary.get("traceLengthFromSources"),
                storedfilter_instance.traceLengthFromSources,
            )
            self.assertIsInstance(
                storedfilter_instance.traceLengthFromSources,
                list,
            )

        if test_traceLengthToSinks:
            self.assertEqual(
                filter_dictionary.get("traceLengthToSinks"),
                storedfilter_instance.traceLengthToSinks,
            )
            self.assertIsInstance(
                storedfilter_instance.traceLengthToSinks,
                list,
            )

        if test_is_new_issue:
            self.assertEqual(
                filter_dictionary.get("is_new_issue"),
                storedfilter_instance.is_new_issue,
            )
            self.assertIsInstance(
                storedfilter_instance.is_new_issue,
                bool,
            )

    def test_instantiate_storedfilter(self) -> None:
        filter_json = json.loads(json.dumps(filter_dictionary))
        storedfilter_instance = StoredFilter(
            filter_json.pop("name"), filter_json.pop("description"), **filter_json
        )
        self._test_filter_equality(storedfilter_instance)

    def test_instantiate_storedfilter_with_kwargs(self) -> None:
        filter_json = json.loads(json.dumps(filter_dictionary))
        storedfilter_instance = StoredFilter(**filter_json)
        self._test_filter_equality(storedfilter_instance)

    def test_partial_instantiate_storedfilter(self) -> None:
        incomplete_filter_json = json.loads(json.dumps(filter_dictionary))
        incomplete_filter_json.pop("description")
        incomplete_filter_json.pop("paths")
        incomplete_filter_json.pop("callables")
        incomplete_filter_json.pop("traceLengthToSinks")
        incomplete_filter_json.pop("is_new_issue")
        partial_storedfilter_instance = StoredFilter(**incomplete_filter_json)
        self._test_filter_equality(
            partial_storedfilter_instance,
            test_description=False,
            test_paths=False,
            test_callables=False,
            test_traceLengthToSinks=False,
            test_is_new_issue=False,
        )
        self.assertEqual("", partial_storedfilter_instance.description)
        self.assertEqual([], partial_storedfilter_instance.paths)
        self.assertEqual([], partial_storedfilter_instance.callables)
        self.assertEqual(None, partial_storedfilter_instance.traceLengthToSinks)
        self.assertEqual(None, partial_storedfilter_instance.is_new_issue)

    def test_fails_instantiate_storedfilter_without_name(self) -> None:
        filter_json = json.loads(json.dumps(filter_dictionary))
        filter_json.pop("name")
        with self.assertRaises(TypeError):
            StoredFilter(**filter_json)
        with self.assertRaises(FilterValidationException):
            # pyre-ignore[6] This is intended for the test
            StoredFilter(name=None, **filter_json)

    def test_fails_instantiate_storedfilter_without_description(self) -> None:
        filter_json = json.loads(json.dumps(filter_dictionary))
        filter_json.pop("description")
        self.assertEqual("", StoredFilter(**filter_json).description)
        with self.assertRaises(FilterValidationException):
            # pyre-ignore[6] This is intended for the test
            StoredFilter(description=None, **filter_json)

    def test_fails_instantiate_storedfilter_without_filtering_conditions(self) -> None:
        filter_json = json.loads(json.dumps(filter_dictionary))
        filter_json.pop("features")
        filter_json.pop("codes")
        filter_json.pop("paths")
        filter_json.pop("callables")
        filter_json.pop("traceLengthFromSources")
        filter_json.pop("traceLengthToSinks")
        filter_json.pop("is_new_issue")
        with self.assertRaises(FilterValidationException):
            StoredFilter(**filter_json)

    def test_storedfilter_json_filtering_keys(self) -> None:
        filter_json = json.loads(json.dumps(filter_dictionary))
        self.assertEqual(
            [
                "features",
                "codes",
                "paths",
                "callables",
                "statuses",
                "source_names",
                "source_kinds",
                "sink_names",
                "sink_kinds",
                "traceLengthFromSources",
                "traceLengthToSinks",
                "is_new_issue",
            ],
            StoredFilter(**filter_json)._json_filtering_keys(),
        )

    def test_storedfilter_to_json(self) -> None:
        filter_json = json.loads(json.dumps(filter_dictionary))
        expected_json_output = json.loads(json.dumps(filter_dictionary))
        expected_json_output.pop("name")
        expected_json_output.pop("description")
        self.assertEqual(
            json.dumps(expected_json_output),
            StoredFilter(**filter_json).to_json(),
        )

    def test_partial_storedfilter_to_json(self) -> None:
        incomplete_filter_json = json.loads(json.dumps(filter_dictionary))
        incomplete_filter_json.pop("description")
        incomplete_filter_json.pop("paths")
        incomplete_filter_json.pop("callables")
        incomplete_filter_json.pop("traceLengthToSinks")
        incomplete_filter_json.pop("is_new_issue")
        expected_json_output = json.loads(json.dumps(incomplete_filter_json))
        expected_json_output.pop("name")
        self.assertEqual(
            json.dumps(expected_json_output),
            StoredFilter(**incomplete_filter_json).to_json(),
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import logging
import re
import sys
from typing import (
    IO,
    Any,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Set,
    Tuple,
    Union,
)

from .. import pipeline as sapp
from ..analysis_output import AnalysisOutput, Metadata
from .base_parser import BaseParser

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


LOG: logging.Logger = logging.getLogger()
UNKNOWN_PATH: str = "unknown"
UNKNOWN_LINE: int = -1


# NOTE: This may or may not produce desired results if there is a number
# in the field name; need to find an example
def _upper_camel_case_to_snake_case(string: str) -> str:
    return re.sub("([a-z])([A-Z])", "\\1_\\2", string).lower()


class Method(NamedTuple):
    name: str

    def is_leaf(self) -> bool:
        return self.name == "leaf"

    @staticmethod
    def from_json(method: Union[None, str, Dict[str, Any]]) -> "Method":
        if method is None:
            return Method("leaf")
        if isinstance(method, str):
            return Method(method)

        canonical_name = method["name"]

        parameter_type_overrides = method.get("parameter_type_overrides")
        if parameter_type_overrides:
            parameter_type_overrides = (
                f"{override['parameter']}: {override['type']}"
                for override in parameter_type_overrides
            )
            canonical_name += "[%s]" % ", ".join(parameter_type_overrides)

        return Method(canonical_name)


class Port(NamedTuple):
    value: str

    def is_leaf(self) -> bool:
        return (
            self.value in ("source", "sink")
            or self.value.startswith("anchor:")
            or self.value.startswith("producer:")
        )

    @staticmethod
    def to_crtex(port: str) -> str:
        """Converts 'argument(n)' to 'formal(n)'. Other CRTEX tools use 'formal'
        to denote argument positions."""
        return re.sub(r"argument\((-?\d+)\)", r"formal(\1)", port)

    @staticmethod
    def from_json(port: str, leaf_kind: str) -> "Port":
        elements = port.split(".")

        if len(elements) == 0:
            raise sapp.ParseError(f"Invalid port: `{port}`.")

        elements[0] = elements[0].lower()
        if elements[0] == "leaf":
            elements[0] = leaf_kind
        elif elements[0] == "return":
            elements[0] = "result"
        elif elements[0] == "anchor":
            # Anchor port is of the form Anchor.<MT port, e.g. Argument(0)>
            # SAPP/CRTEX expects: "anchor:formal(0)"
            canonical_port = Port.from_json(
                ".".join(elements[1:]), "unreachable_leaf_kind_anchor"
            )
            return Port(f"{elements[0]}:{Port.to_crtex(canonical_port.value)}")
        elif elements[0] == "producer" and len(elements) >= 3:
            # Producer port is of the form Producer.<producer_id>.<MT port>.
            # SAPP/CRTEX expects: "producer:<producer_id>:<canonical_port>".
            root = elements[0]
            producer_id = elements[1]
            canonical_port = Port.from_json(
                ".".join(elements[2:]), "unreachable_leaf_kind_producer"
            )
            return Port(f"{root}:{producer_id}:{Port.to_crtex(canonical_port.value)}")

        return Port(".".join(elements))


class Position(NamedTuple):
    path: str
    line: int
    start: int
    end: int

    @staticmethod
    def from_json(position: Dict[str, Any], method: Method) -> "Position":
        path = position.get("path", UNKNOWN_PATH)
        line = position.get("line", UNKNOWN_LINE)
        start = position.get("start", 0) + 1
        end = max(position.get("end", 0) + 1, start)
        if path == UNKNOWN_PATH and not method.is_leaf():
            path = method.name.split(";")[0]
            path = path.split("$")[0]
            path = path[1:]
        return Position(path, line, start, end)

    def to_sapp(self) -> sapp.SourceLocation:
        return sapp.SourceLocation(
            line_no=self.line,
            begin_column=self.start,
            end_column=self.end,
        )


class Call(NamedTuple):
    method: Method
    port: Port
    position: Position

    @staticmethod
    def from_json(
        method: Union[None, str, Dict[str, Any]],
        port: str,
        position: Optional[Dict[str, Any]],
        default_position: Position,
        leaf_kind: str,
    ) -> "Call":
        call_method = Method.from_json(method)
        call_port = Port.from_json(port, leaf_kind)
        if position is None:
            if not call_port.is_leaf():
                raise sapp.ParseError(
                    f"Missing call position for call to `{call_method.name}`."
                )
            else:
                call_position = default_position
        else:
            call_position = Position.from_json(position, call_method)
        return Call(call_method, call_port, call_position)


class LocalPositions(NamedTuple):
    positions: List[Position]

    @staticmethod
    def from_json(positions: List[Dict[str, Any]], method: Method) -> "LocalPositions":
        return LocalPositions(
            [Position.from_json(position, method) for position in positions]
        )

    def to_sapp(self) -> List[sapp.SourceLocation]:
        return [position.to_sapp() for position in sorted(self.positions)]


class Features(NamedTuple):
    features: Set[str]

    @staticmethod
    def from_json(features: Dict[str, Any]) -> "Features":
        may_features = set(
            features.get("may_features", []) + features.get("always_features", [])
        )
        always_features = {
            f"always-{feature}" for feature in features.get("always_features", [])
        }
        return Features(may_features | always_features)

    def to_sapp(self) -> List[str]:
        return sorted(self.features)


class Condition(NamedTuple):
    caller: Call
    callee: Call
    kind: str
    distance: int
    local_positions: LocalPositions
    features: Features

    def convert_to_sapp(
        self, kind: Literal[sapp.ParseType.PRECONDITION, sapp.ParseType.POSTCONDITION]
    ) -> sapp.ParseConditionTuple:
        return sapp.ParseConditionTuple(
            type=kind,
            caller=self.caller.method.name,
            caller_port=self.caller.port.value,
            filename=self.caller.position.path,
            callee=self.callee.method.name,
            callee_port=self.callee.port.value,
            callee_location=self.callee.position.to_sapp(),
            type_interval=None,
            features=self.features.to_sapp(),
            titos=self.local_positions.to_sapp(),
            leaves=[(self.kind, self.distance)],
            annotations=[],
        )


class Precondition(Condition):
    def to_sapp(self) -> sapp.ParseConditionTuple:
        return super().convert_to_sapp(sapp.ParseType.PRECONDITION)


class Postcondition(Condition):
    def to_sapp(self) -> sapp.ParseConditionTuple:
        return super().convert_to_sapp(sapp.ParseType.POSTCONDITION)


class IssueCondition(NamedTuple):
    callee: Call
    kind: str
    distance: int
    local_positions: LocalPositions
    features: Features

    def to_sapp(self) -> sapp.ParseIssueConditionTuple:
        return sapp.ParseIssueConditionTuple(
            callee=self.callee.method.name,
            port=self.callee.port.value,
            location=self.callee.position.to_sapp(),
            leaves=[(self.kind, self.distance)],
            titos=self.local_positions.to_sapp(),
            features=self.features.to_sapp(),
            type_interval=None,
            annotations=[],
        )


class Leaf(NamedTuple):
    method: Method
    kind: str
    distance: int

    def to_sapp(self) -> sapp.ParseIssueLeaf:
        return (self.method.name, self.kind, self.distance)


class Issue(NamedTuple):
    code: int
    message: str
    callable: Method
    callable_position: Position
    issue_position: Position
    preconditions: List[IssueCondition]
    postconditions: List[IssueCondition]
    initial_sources: Set[Leaf]
    final_sinks: Set[Leaf]
    features: Features

    def to_sapp(self, parser: "Parser") -> sapp.ParseIssueTuple:
        return sapp.ParseIssueTuple(
            code=self.code,
            message=self.message,
            callable=self.callable.name,
            handle=parser.compute_master_handle(
                callable=self.callable.name,
                line=self.issue_position.line - self.callable_position.line,
                start=self.issue_position.start,
                end=self.issue_position.end,
                code=self.code,
            ),
            filename=self.callable_position.path,
            callable_line=self.callable_position.line,
            line=self.issue_position.line,
            start=self.issue_position.start,
            end=self.issue_position.end,
            preconditions=[
                precondition.to_sapp() for precondition in self.preconditions
            ],
            postconditions=[
                postcondition.to_sapp() for postcondition in self.postconditions
            ],
            initial_sources={leaf.to_sapp() for leaf in self.initial_sources},
            final_sinks={leaf.to_sapp() for leaf in self.final_sinks},
            features=self.features.to_sapp(),
            fix_info=None,
        )


class Parser(BaseParser):
    def __init__(self, repo_dirs: Optional[List[str]] = None) -> None:
        super().__init__(repo_dirs)
        # pyre-fixme[4]: Attribute annotation cannot contain `Any`.
        self._rules: Dict[int, Any] = {}
        self._initialized: bool = False

    @staticmethod
    def is_supported(metadata: Metadata) -> bool:
        return metadata.tool == "mariana_trench"

    def initialize(self, metadata: Optional[Metadata]) -> None:
        if self._initialized:
            return

        if metadata:
            # We get the rules from the metadata when parsing json lines.
            rules = metadata.rules
            if rules:
                self._rules = rules
                self._initialized = True

    # pyre-fixme[14]: `parse` overrides method defined in `BaseParser` inconsistently.
    def parse(
        self, output: AnalysisOutput
    ) -> Iterable[Union[sapp.ParseConditionTuple, sapp.ParseIssueTuple]]:
        self.initialize(output.metadata)

        for handle in output.file_handles():
            yield from self.parse_handle(handle)

    def parse_handle(
        self, handle: IO[str]
    ) -> Iterable[Union[sapp.ParseConditionTuple, sapp.ParseIssueTuple]]:
        for line in handle.readlines():
            if line.startswith("//"):
                continue
            model = json.loads(line)
            yield from self._parse_issues(model)
            for precondition in self._parse_precondition(model):
                yield precondition.to_sapp()
            for postcondition in self._parse_postconditions(model):
                yield postcondition.to_sapp()

    def _parse_issues(self, model: Dict[str, Any]) -> Iterable[sapp.ParseIssueTuple]:
        for issue in model.get("issues", []):
            code = issue["rule"]
            rule = self._rules[code]
            callable = Method.from_json(model["method"])
            callable_position = Position.from_json(model["position"], callable)
            issue_position = Position.from_json(issue["position"], callable)
            features = Features.from_json(issue)

            (preconditions, final_sinks) = self._parse_issue_conditions(
                issue, callable, callable_position, "sink"
            )
            (postconditions, initial_sources) = self._parse_issue_conditions(
                issue, callable, callable_position, "source"
            )

            yield Issue(
                code=code,
                message=f"{rule['name']}: {rule['description']}",
                callable=callable,
                callable_position=callable_position,
                issue_position=issue_position,
                preconditions=preconditions,
                postconditions=postconditions,
                initial_sources=initial_sources,
                final_sinks=final_sinks,
                features=features,
            ).to_sapp(self)

    def _parse_issue_conditions(
        self,
        issue: Dict[str, Any],
        callable: Method,
        callable_position: Position,
        leaf_kind: str,
    ) -> Tuple[List[IssueCondition], Set[Leaf]]:
        frames = issue[f"{leaf_kind}s"]

        conditions = []
        leaves = set()

        for frame in frames:
            frames = Parser._normalize_frame(frame)
            for frame in frames:
                conditions.append(
                    IssueCondition(
                        callee=Call.from_json(
                            method=frame.get("callee"),
                            port=frame["callee_port"],
                            position=frame.get("call_position"),
                            default_position=callable_position,
                            leaf_kind=leaf_kind,
                        ),
                        kind=frame["kind"],
                        distance=frame.get("distance", 0),
                        local_positions=LocalPositions.from_json(
                            frame.get("local_positions", []), callable
                        ),
                        features=Features.from_json(frame.get("local_features", {})),
                    )
                )

                for origin in frame.get("origins", []):
                    leaves.add(
                        Leaf(
                            method=Method.from_json(origin),
                            kind=frame["kind"],
                            distance=frame.get("distance", 0),
                        )
                    )

        return conditions, leaves

    @staticmethod
    def _normalize_frame(frame: Dict[str, Any]) -> List[Dict[str, Any]]:
        if "canonical_names" not in frame:
            return [frame]

        frames = []
        # Expected format: "canonical_names": [ { "instantiated": "<name>" }, ... ]
        # Canonical names are used for CRTEX only, and are expected to be the callee name
        # where traces are concerened. Each instantiated name maps to one frame.
        for canonical_name in frame["canonical_names"]:
            if "instantiated" not in canonical_name:
                # Uninstantiated canonical names are user-defined CRTEX leaves
                # They do not show up as a frame in the UI.
                continue

            frame_copy = frame.copy()  # Shallow copy is ok, only "callee" is different.
            frame_copy["callee"] = canonical_name["instantiated"]
            frames.append(frame_copy)

        return frames

    def _parse_precondition(self, model: Dict[str, Any]) -> Iterable[Precondition]:
        caller = Method.from_json(model["method"])
        caller_position = Position.from_json(model["position"], caller)

        for sink in model.get("sinks", []):
            sinks = Parser._normalize_frame(sink)
            for sink in sinks:
                yield Precondition(
                    caller=Call(
                        method=caller,
                        port=Port.from_json(sink["caller_port"], "sink"),
                        position=caller_position,
                    ),
                    callee=Call.from_json(
                        method=sink.get("callee"),
                        port=sink["callee_port"],
                        position=sink.get("call_position"),
                        default_position=caller_position,
                        leaf_kind="sink",
                    ),
                    kind=sink["kind"],
                    distance=sink.get("distance", 0),
                    local_positions=LocalPositions.from_json(
                        sink.get("local_positions", []), caller
                    ),
                    features=Features.from_json(sink.get("local_features", {})),
                )

    def _parse_postconditions(self, model: Dict[str, Any]) -> Iterable[Postcondition]:
        caller = Method.from_json(model["method"])
        caller_position = Position.from_json(model["position"], caller)

        for generation in model.get("generations", []):
            generations = Parser._normalize_frame(generation)
            for generation in generations:
                yield Postcondition(
                    caller=Call(
                        method=caller,
                        port=Port.from_json(generation["caller_port"], "source"),
                        position=caller_position,
                    ),
                    callee=Call.from_json(
                        method=generation.get("callee"),
                        port=generation["callee_port"],
                        position=generation.get("call_position"),
                        default_position=caller_position,
                        leaf_kind="source",
                    ),
                    kind=generation["kind"],
                    distance=generation.get("distance", 0),
                    local_positions=LocalPositions.from_json(
                        generation.get("local_positions", []), caller
                    ),
                    features=Features.from_json(generation.get("local_features", {})),
                )
